

import dataclasses


@dataclasses.dataclass
class RuntimeConfig:
    small_text_provider: str = None
    small_text_model: str = None

    medium_text_provider: str = None
    medium_text_model: str = None

    large_text_provider: str = 'openai_llm'
    large_text_model: str = 'gpt-4'

    medium_image_provider: str = None
    medium_image_model: str = None
    medium_image_width: int = None
    medium_image_height: int = None


GLBOAL_RUNTIME_CONFIG = RuntimeConfig()


def resolve_runtime_provider_and_model(runtime_provider, runtime_model):
    if 'RUNTIME:' not in runtime_provider:
        raise NotImplementedError('Only RUNTIME: providers are supported')
    if 'RUNTIME:' not in runtime_model:
        raise NotImplementedError('Only RUNTIME: models are supported')

    runtime_provider = runtime_provider.replace('RUNTIME:', '')
    runtime_model = runtime_model.replace('RUNTIME:', '')

    prov = getattr(GLBOAL_RUNTIME_CONFIG, f'{runtime_provider}')
    model = getattr(GLBOAL_RUNTIME_CONFIG, f'{runtime_model}')
    return prov, model





from collections import namedtuple
from enum import Enum
import time


class EventType(Enum):
    TASK_START = 1
    TASK_SUCCESS = 2
    TASK_FAILURE = 3
    THOUGHT_CRIME = 4
    LLM_INFERENCE_START = 5
    LLM_INFERENCE_END = 6
    RAG_QUERY_START = 7
    RAG_QUERY_END = 8
    CONFIG_DISCOVER = 9
    IMAGE_INFERENCE_START = 10
    IMAGE_INFERENCE_END = 11
    THOUGHT_CRIME_RETRY = 12
    PARAM_THUMBS_DOWN = 13
    PARAM_THUMBS_UP = 14
    PARAM_PARTICIPATION = 15
    PARAM_THOUGHT_CRIME = 16


MAP_START_TO_END = {
    EventType.LLM_INFERENCE_START: EventType.LLM_INFERENCE_END,
    EventType.RAG_QUERY_START: EventType.RAG_QUERY_END,
    EventType.IMAGE_INFERENCE_START: EventType.IMAGE_INFERENCE_END
}


EVENTS = {
    EventType.LLM_INFERENCE_START: namedtuple('LlmInferenceStart', ['model_name', 'temperature', 'max_tokens', 'user_message', 'system_prompt', 'history_pairs', 'provider_name']),
    EventType.LLM_INFERENCE_END: namedtuple('LlmInferenceEnd', ['model_name', 'duration_s', 'input_tokens', 'output_tokens', 'response', 'provider_name']),
    EventType.RAG_QUERY_START: namedtuple('RagQueryStart', ['query', 'k', 'rag_provider']),
    EventType.RAG_QUERY_END: namedtuple('RagQueryEnd', ['duration_s', 'k', 'results', 'rag_provider']),
    EventType.CONFIG_DISCOVER: namedtuple('ConfigDiscover', ['message']),
    EventType.IMAGE_INFERENCE_START: namedtuple('ImageInferenceStart', ['image_prompt', 'image_model_name', 'image_width', 'image_height', 'num_images', 'provider_name']),
    EventType.IMAGE_INFERENCE_END: namedtuple('ImageInferenceEnd', ['image_model_name', 'duration_s', 'image_list', 'image_width', 'image_height', 'provider_name']),
    EventType.THOUGHT_CRIME: namedtuple('ThoughtCrime', ['message']),
    EventType.THOUGHT_CRIME_RETRY: namedtuple('ThoughtCrimeRetry', ['message']),

    # Param Logalytics
    EventType.PARAM_THUMBS_DOWN: namedtuple('ParamThumbsDown', ['param_ref', 'param_ref_versioned']),
    EventType.PARAM_THUMBS_UP: namedtuple('ParamThumbsUp', ['param_ref', 'param_ref_versioned']),
    EventType.PARAM_THOUGHT_CRIME: namedtuple('ParamThoughtCrime', ['param_ref', 'param_ref_versioned'])
}


class Event:
    def __init__(self, name: str, event_type: EventType, fields: dict):
        self.name = name
        self.event_time = time.time()
        self.event_type = event_type
        self.info = EVENTS[event_type](**fields)




import axiomic.protos as protos
import axiomic.errors as errors
import axiomic.engine.node_impls.builtin_unary_checks as builtin_unary_checks 


def builtin_unary_check_node(bucn: protos.axiomic.BuiltinUnaryCheckNode, context, weave_node: protos.axiomic.AxiomicNode) -> str:
    incoming_str = context.resolve_node(bucn.input)
    check_name = bucn.unary_check_name

    if check_name not in builtin_unary_checks.CHECKS:
        raise errors.WeaveError(f'Unknown check: {check_name}')

    answer = builtin_unary_checks.CHECKS[check_name](incoming_str)
    return answer




import grpc
import os

from google.protobuf import text_format as text_format_
from google.protobuf.json_format import MessageToJson



def text_parse(proto, text):
    text_format_.Parse(text, proto)
    return proto

def text_dump(proto):
    return text_format_.MessageToString(proto)

text_format = text_dump

def _load_protos(file_path):
    protos, service = grpc.protos_and_services(os.path.join('axiomic/protos', file_path))
    return protos, service

def json_dump(proto):
    return MessageToJson(proto)

axiomic, axiomic_service = _load_protos('axiomic.proto')



import inspect
import os

import axiomic.providers.default_providers as default_providers

import axiomic.models.context as _context

import axiomic.logalytics.sink as sink

import axiomic.models.models_openai as quick_openai
import axiomic.models.models_together as quick_together
import axiomic.models.models_anthropic as quick_anthropic
import axiomic.models.generic as generic
import axiomic.configure.runtime_config as runtime_config


Generic = generic.Generic


def discover_user_calling_file():
    frame = inspect.currentframe()
    weave_package_name = "axiomic" 

    while frame:
        module_name = frame.f_globals.get("__name__")
        if module_name and not module_name.startswith(weave_package_name):
            calling_file = frame.f_code.co_filename
            return os.path.abspath(calling_file)
        frame = frame.f_back
    return None


def _discovery(env_key, prov):
    if env_key not in os.environ:
        return
    _context.register_global_provider(prov)
    _context.GLOBAL_DEFAULT_CONTEXT.update_non_none(**prov.get_default_context_params())


def _announce_provider(env_key, models_name):
    if env_key in os.environ:
        sink.discover_config(f'discovered {env_key}, models available under models.{models_name}')


def discover_providers():
    _context.register_global_ref_provider(default_providers.FILESYSTME_PARAMS)

    _announce_provider('OPENAI_API_KEY', 'OpenAI')
    _announce_provider('ANTHROPIC_API_KEY', 'Anthropic')
    _announce_provider('TOGETHER_API_KEY', 'Together')

    # Open AI discovery
    _discovery('OPENAI_API_KEY', default_providers.OPENAI_LLM)
    _discovery('OPENAI_API_KEY', default_providers.OPENAI_IMG)
    _discovery('OPENAI_API_KEY', default_providers.OPENAI_EMBED)

    if 'OPENAI_API_KEY' in os.environ:
        quick_openai.OpenAI.bind()

    # Anthropic Discovery
    _discovery('ANTHROPIC_API_KEY', default_providers.ANTHROPIC_LLM)

    if 'ANTHROPIC_API_KEY' in os.environ:
        quick_anthropic.Anthropic.bind()

    # Together Discovery
    _discovery('TOGETHER_API_KEY', default_providers.TOGETHER_LLM)
    _discovery('TOGETHER_API_KEY', default_providers.TOGETHER_IMG)

    if 'TOGETHER_API_KEY' in os.environ:
        quick_together.Together.bind()



import axiomic.protos as protos

import axiomic.providers.img_provider.img_provider as img_provider
import axiomic.configure as configure

from icecream import ic 

ic.configureOutput(includeContext=True)


def txt_to_img_node(node: protos.axiomic.TextToImagesNode, c, weave_node: protos.axiomic.AxiomicNode) -> str:
    prompt = c.resolve_node(node.image_prompt)

    params = node.params
    req = img_provider.ImgInferenceRequest(
        params.image_provider_name,
        params.model_name,
        params.image_width,
        params.image_height,
        prompt,
        params.num_images
    )

    resp = configure.get_provider_provider().get_provider(params.image_provider_name).infer(req)
    return resp.image_list



import axiomic.frontend.config as _config
import axiomic.frontend.text as _text
import axiomic.frontend.image as _image
import axiomic.errors as _errors
import axiomic.frontend.params as _params
import axiomic.frontend.functions as _functions

import axiomic.models as models



# axiomic operations
Text = _text.Text
TextType = _text.TextType

# Images
Image = _image.Image


# Errors
ThoughtCrime = _errors.ThoughtCrime
GraphError = _errors.GraphError


# Functions
infer = _functions.infer
format = _functions.format


# Params
P = _params.P



import axiomic.data.lists as lists

import axiomic.models as wc
import collections

import json
import yaml
import jsonschema
import pydantic

import axiomic.data.validate as validate


LIST_OF_STRINGS_SCHEMA = {'type': 'array', 'items': {'type': 'string'}}
LIST_SCHEMA = {'type': 'array'}

CHAT_SCHEMA = {
    "type": "array",
    "items": {
        "type": "object",
        "properties": {
            "user": {
                "type": "string"
            },
            "agent": {
                "type": "string"
            }
        },
        "required": ["user", "agent"]
    }
}

CHAT_SCHEMA_STRUCT = {
    "type": "array",
    "items": {
        "type": "object",
        "required": ["user", "agent"]
    }
}


def _read(filename: str) -> any:
    filename = wc.resolve_axiomic_variables(filename)
    if '$' in filename:
        raise ValueError(f'Failed to resolve axiomic variables in filename: {filename}')

    with open(filename, 'r') as f:
        return f.read()


def read_text(filename: str) -> str:
    '''
    Reads a file and returns the contents as a string.

    Only reads .txt files.

    Will do axiomic variable expansion in the filename,
    e.g. `read_text('$axiomic_DATA_ROOT/file.txt')`.

    Args:
        filename (str): The filename to read.

    Returns:
        The contents of the file as a string.
    '''
    contents = _read(filename)
    # fail if structured
    if filename.endswith('.json') or filename.endswith('.yaml'):
        raise ValueError(f'Expected text file (.txt): {filename}')
    return contents


def read_structured(filename: str) -> any:
    contents = _read(filename)

    if filename.endswith('.json'):
        return json.loads(contents)
    if filename.endswith('.yaml'):
        return yaml.load(contents, Loader=yaml.FullLoader)
    raise ValueError(f'Unsupported file type (execpted .json, .yaml): {filename}')


def _check_schema(filename, data, schema):
    jsonschema.validate(instance=data, schema=schema)


def load_strings(filename: str) -> lists.TextList:
    '''
    Loads a list of strings from a file.


    Example, `file.json` contains:
    ```
        [
            "one",
            "two"
        ]
    ```
    Or `file.yaml` contains:
    ```
        - one
        - two
    ```

    Loads as this structure:
    ```
        axiomic.data.TextList(["one", "two"])
    ```

    Args:
        filename (str): The filename to load.

    Returns:
        A TextList of strings.
    '''
    result = read_structured(filename)
    _check_schema(filename, result, LIST_OF_STRINGS_SCHEMA)
    return lists.TextList(result)


def load_jsons_parsed(filename: str, schema: str = None) -> lists.TextList:
    '''
    Loads a file as a TextList of jsons. The JSONs are deserialized.

    Can load either a YAML or JSON file.

    Example, `file.json` contains:
    ```
        [
            {"a": 1, "b": 2},
            {"a": 3, "b": 4}]
        ]
    ```
    Or `file.yaml` contains:
    ```
        - a: 1
          b: 2
        - a: 3
          b: 4
    ```

    Loads as this structure:
    ```
        axiomic.data.TextList([{"a": 1, "b": 2}, {"a": 3, "b": 4}])
    ```

    Parameters:
        filename (str): The filename to load.
        schema (str): Optional, the schema to validate the JSONs against.

    Returns:
        A TextList of unserialize JSON objects.
    '''
    result = read_structured(filename)
    if schema is not None:
        _check_schema(filename, result, LIST_SCHEMA)
    return lists.TextList(result)


def load_jsons_unparsed(filename: str, schema: str = None) -> lists.TextList:
    '''
    Loads a file as a TextList of jsons. The JSONs are not deserialized.

    Can load either a YAML or JSON file.

    Example, `file.json` contains:
    ```
        [
            {"a": 1, "b": 2},
            {"a": 3, "b": 4}]
        ]
    ```
    Or `file.yaml` contains:
    ```
        - a: 1
          b: 2
        - a: 3
          b: 4
    ```

    Loads as this structure:
    ```
        axiomic.data.TextList(['{"a": 1, "b": 2}', '{"a": 3, "b": 4}'])
    ```

    Parameters:
        filename (str): The filename to load.
        schema (str): Optional, the schema to validate the JSONs against.

    Returns:
        A TextList of serialized JSON objects.
    '''
    result = read_structured(filename)
    if schema is not None:
        _check_schema(filename, result, schema)
    l = []
    for s in result:
        l.append(json.dumps(s))
    return lists.TextList(l)


def load_pydantics(filename: str, pydantic_model: pydantic.BaseModel) -> lists.TextList:
    jsons = load_jsons_parsed(filename)

    pydantics = []
    for j in jsons:
        pydantics.append(pydantic_model(**j))

    return lists.TextList(pydantics)


def load_chat_text(filename: str, strip=True) -> lists.ChatList:
    '''
    Loads a chat from a file.

    Example, `file.json` contains:
    ```
        [
            {"user": "Hello Agent!", "agent": "Hi Person!"},
            {"user": "What's 5 + 5?", "agent": "It's 10"},
        ]
    ```
    Or `file.yaml` contains:
    ```
        - user: Hello Agent!
          agent: Hi Person!
        - user: What's 5 + 5?
          agent: It's 10
    ```

    Loads as this structure:
    ```
        axiomic.data.ChatList([("Hello Agent!", "Hi Person!"), ("What's 5 + 5?", "It's 10")])
    ```

    Args:
        filename (str): The filename to load.
        strip (bool): If True, strip whitespace from the user and agent strings.

    Returns:
        A axiomic.data.ChatList.
    '''
    result = read_structured(filename)
    _check_schema(filename, result, CHAT_SCHEMA)
    pairs = [(d['user'], d['agent']) for d in result]

    if strip:
        pairs = [(user.strip(), agent.strip()) for user, agent in pairs]

    return lists.ChatList(pairs)


def load_chat_jsons_unparsed(filename: str, user_schema: validate.JsonSchemaType, agent_schema: validate.JsonSchemaType, strip=True) -> lists.ChatList:
    '''
    Loads a chat from file, validating if the user and agent strings are valid JSON.

    Example, if `file.json` contains:
    ```
        [
            {"user": {"a": 1}, "agent": {"b": 2}},
            {"user": {"a": 2}, "agent": {"b": 3}},
        ]
    ```
    Becomes:
    ```
        axiomic.data.ChatList([('{"a": 1}', '{"b": 2}'), ('{"a": 2}', '{"b": 3}')]
    ```

    Args:
        filename (str): The filename to load.
        user_schema (any): A custom validator to validate the user strings.
        agent_schema (any): A custom validator to validate the agent strings.

    Returns:
        A axiomic.data.ChatList.
    '''
    parsed = load_chat_jsons_parsed(filename, user_schema, agent_schema, strip)
    unparsed = []
    for chat in parsed:
        unparsed.append((json.dumps(chat.user), json.dumps(chat.agent)))
    return lists.ChatList(unparsed)


def load_chat_jsons_parsed(filename: str, user_schema: validate.JsonSchemaType, agent_schema: validate.JsonSchemaType, strip=True) -> lists.ChatList:
    read = read_structured(filename)
    _check_schema(filename, read, CHAT_SCHEMA_STRUCT)

    def strip_if_string(text):
        if not strip:
            return text
            
        if isinstance(text, str):
            return text.strip()
        return text

    pairs = [(strip_if_string(d['user']), strip_if_string(d['agent'])) for d in read]
    serialized = []
    user_validator = validate.ValidatorJson(user_schema)
    agent_validator = validate.ValidatorJson(agent_schema)
    for user, agent in pairs:
        new_user = None
        new_agent = None
        user_validator.validate(user)
        agent_validator.validate(agent)
        serialized.append((user, agent))

    return lists.ChatList(serialized)


def load_chat_pydantic(filename: str, user_model: pydantic.BaseModel, agent_model: pydantic.BaseModel) -> lists.ChatList:
    user_schema = user_model.schema()
    agent_schema = agent_model.schema()
    parsed_chat = load_chat_jsons_parsed(filename, user_schema, agent_schema)

    pairs = []

    for c in parsed_chat:
        u = user_model(**c.user)
        a = agent_model(**c.agent)
        pairs.append((u, a))
    
    return lists.ChatList(pairs)





import axiomic.protos as protos

from icecream import ic 


ic.configureOutput(includeContext=True)


def flat_rag_topk_node(node: protos.axiomic.FlatRagTopKNode, c, weave_node: protos.axiomic.AxiomicNode) -> str:
    query = c.resolve_node(node.query)
    separator = c.resolve_node(node.separator)
    rag_provider = c.get_provider(node.rag_provider)
    
    result_list = rag_provider.query_topk(query, k=node.k)
    result_str = separator.join(result_list)
    return result_str




import json
import typing
import jsonschema
import axiomic.frontend.text as text


JsonSchemaType = typing.Union[str, dict]


def _to_schema_dict(schema: JsonSchemaType):
    if isinstance(schema, dict):
        return schema
    # Pass through parameter expansion
    return json.loads(Text(schema).value())


class ValidatorString:
    def validate(self, thing: any) -> bool:
        return isinstance(thing, str)


class ValidateSerializedJson:
    def validate(self, thing: any) -> bool:
        try:
            json.loads(thing)
            return True
        except json.JSONDecodeError:
            return False


class ValidatorJson:
    def __init__(self, schema: typing.Union[str, dict]):
        self.schema = _to_schema_dict(schema)
        
    def validate(self, thing: any) -> bool:
        try:
            jsonschema.validate(thing, self.schema)
            return True
        except jsonschema.ValidationError:
            return False
        except jsonschema.SchemaError:
            return False


class ValidatorJsonSchemaSerialized:
    def __init__(self, schema: typing.Union[str, dict]):
        self.schema = _to_schema_dict(schema)
        
    def validate(self, thing: str) -> bool:
        try:
            jsonschema.validate(json.loads(thing), self.schema)
            return True
        except jsonschema.ValidationError:
            return False
        except jsonschema.SchemaError:
            return False


class ValidatorPydantic:
    def __init__(self, pydantic_model: any):
        self.pydantic_model = pydantic_model
        
    def validate(self, thing: any) -> bool:
        return isinstance(thing, self.pydantic_model)


class Validator:
    def __init__(self, validator: any):
        self.validator = validator





from rich import print
from rich.status import Status
from contextlib import contextmanager

from rich.console import Console, Group
from rich.panel import Panel
from rich.text import Text
from rich import box
from rich.layout import Layout


import axiomic.logalytics.events as events
import axiomic.configure as configure
import axiomic.configure.default_config as default_config
# import axiomic.models as models


@contextmanager
def spinner(message="Processing...", spinner="aesthetic"):
    status = Status(message, spinner=spinner)
    status.start()
    try:
        yield
    finally:
        status.stop()

def start_spinner(message="Processing..."):
    global spinner_context
    spinner_context = spinner(message=message)
    spinner_context.__enter__()

def stop_spinner():
    spinner_context.__exit__(None, None, None)


def verbose_llm_inference_end(e: events.Event):
    # 'model_name', 'duration_s', 'input_tokens', 'output_tokens', 'response'
    model_name = e.info.model_name
    duration_s = e.info.duration_s
    input_tokens = e.info.input_tokens
    output_tokens = e.info.output_tokens
    response = e.info.response
    console = Console()

    footer_text = Text(f"Model: {model_name} | Duration: {duration_s:.2f}s | Input Tokens: {input_tokens} | Output Tokens: {output_tokens}", style="bold")

    # Main panel that includes the layout and footer
    main_panel = Panel(response, title="LLM Inference - Response", style='green', subtitle=footer_text, box=box.ROUNDED, expand=True)

    # Print the main panel
    console.print(main_panel)

    

def verbose_llm_inference(e: events.Event):
    model_name = e.info.model_name
    history_pairs = e.info.history_pairs
    temperature = e.info.temperature
    max_tokens = e.info.max_tokens
    user_message = e.info.user_message
    system_prompt = e.info.system_prompt

    console = Console()

    panel_list = []
    # Create a sub-panel for the system prompt
    if system_prompt:
        panel_list.append(Panel(system_prompt, title="System Prompt")) #, expand=False))

    for i, pair in enumerate(history_pairs):
        panel_list.append(Panel(pair[0], style="dim", title=f"History #{i} - User"))
        panel_list.append(Panel(pair[1], style="dim", title=f"History #{i} - Assistant"))
    # Create sub-panels for history pairs
    # history_panels = [Panel(pair, expand=False) for pair in history_pairs]

    # Create a sub-panel for the user message
    panel_list.append(Panel(user_message, title="User Prompt"))

    # Layout to hold all sub-panels

    panel_group = Group(*panel_list)

    # Footer information as a Text object
    footer_text = Text(f"Model: {model_name} | Temperature: {temperature} | Max Tokens: {max_tokens}", style="bold")
    
    # Main panel that includes the layout and footer
    main_panel = Panel(panel_group, title="LLM Inference - Request", style='dark_green', subtitle=footer_text, box=box.ROUNDED, expand=True)

    # Print the main panel
    console.print(main_panel)


class ConsoleLog:
    def __init__(self):
        pass

    def log_event(self, event: events.Event):
        if event.event_type == events.EventType.LLM_INFERENCE_END:
            self.log_llm_inference_end(event)
        elif event.event_type == events.EventType.LLM_INFERENCE_START:
            self.log_llm_inference_start(event)
        elif event.event_type == events.EventType.RAG_QUERY_START:
            self.rag_query_start(event)
        elif event.event_type == events.EventType.RAG_QUERY_END:
            self.rag_query_end(event)
        elif event.event_type == events.EventType.CONFIG_DISCOVER:
            self.config_discover(event)
        elif event.event_type == events.EventType.IMAGE_INFERENCE_START:
            self.log_image_inference_start(event)
        elif event.event_type == events.EventType.IMAGE_INFERENCE_END:
            self.log_image_inference_end(event)
        else:
            print(f'{event.name} :: {event.info}')

    def log_image_inference_start(self, event: events.Event):
        start_spinner(message=f'Image Inference: {event.name} {event.info.image_model_name} -> {event.info.image_width}x{event.info.image_height}')

    def log_image_inference_end(self, event: events.Event):
        stop_spinner()
        print(f'[bold][blue]{event.name}: {event.info.image_model_name} -> {event.info.image_width}x{event.info.image_height} (in {event.info.duration_s:.2f}s)[/blue][/bold]')

    def config_discover(self, event: events.Event):
        print(f'[dim](axiomic) {event.info.message}[/dim]')

    def log_llm_inference_start(self, event: events.Event):
        # print(f'{event.name} :: {event.info.model_name}')
        # if models.infer_context().verbose_llm:
        #    verbose_llm_inference(event)
        start_spinner(message=f'LLM Inference: {event.name} {event.info.model_name}')

    def log_llm_inference_end(self, event: events.Event):
        stop_spinner()
        print(f'[bold][blue]{event.name}: {event.info.input_tokens} toks -> {event.info.model_name} -> {event.info.output_tokens} toks (in {event.info.duration_s:.2f}s)[/blue][/bold]')
        # if configure.infer_context().verbose_llm:
        #    verbose_llm_inference_end(event)

    def rag_query_start(self, event: events.Event):
        start_spinner(message=f'RAG: {event.info.rag_provider} k={event.info.k}')
    
    def rag_query_end(self, event: events.Event):
        stop_spinner()
        print(f'[bold][dark_green]RAG: {event.info.rag_provider} k={event.info.k} -> {len(event.info.results)} results (in {event.info.duration_s:.2f}s)[/bold][/dark_green]')




import rich
from rich.console import Console
from rich.table import Table


import axiomic.logalytics.events as events


class StatsCounter:
    def __init__(self):
        # Model -> (Count, Duration_s, Input Tokens, Output Tokens)
        self.models = {}

    def print_stats(self):
        if len(self.models) == 0:
            #rich.print()
            #rich.print('[dim]No LLM Inferences[/dim]')
            return

        console = Console()

        table = Table(show_header=True, header_style="dim", box=None)
        table.add_column("Provider", style="dim")
        table.add_column("Model", style="dim")
        table.add_column("Type", justify="right", style="dim")
        table.add_column("Reqs", justify="right", style="dim")
        table.add_column("Time (s)", justify="right", style="dim")
        table.add_column("Tok In", justify="right", style="dim")
        table.add_column("Tok Out", justify="right", style="dim")
        table.add_column("Out/s", justify="right", style="dim")

        rows = []
        for model in sorted(self.models.keys()):
            count, duration, input_tokens, output_tokens, provider_name = self.models[model]
            rate = '-'
            type_ = 'text'
            if duration > 0 and output_tokens > 0:
                rate = f'{output_tokens / duration:.2f}'
            elif duration > 0 and count > 0:
                rate = f'{count / duration:.2f}'
            if input_tokens < 0:
                input_tokens = '-'
                type_ = 'image'
            if output_tokens < 0:
                output_tokens = '-'
            rows.append((provider_name, model, type_, str(count), f'{duration:.2f}', str(input_tokens), str(output_tokens), f'{rate}'))
        
        for row in sorted(rows, key=lambda x: (x[1], x[0])):
            table.add_row(*row)
            

        console.print("\n")
        console.print(table)

    def record_image_inference(self, event: events.Event):
        if event.info.image_model_name not in self.models:
            self.models[event.info.image_model_name] = [0, 0, 0, 0, '']
        
        self.models[event.info.image_model_name][0] += 1
        self.models[event.info.image_model_name][1] += event.info.duration_s
        self.models[event.info.image_model_name][2] = -1
        self.models[event.info.image_model_name][3] = -1
        self.models[event.info.image_model_name][4] = event.info.provider_name

    def log_event(self, event: events.Event):
        if event.event_type == events.EventType.IMAGE_INFERENCE_END:
            self.record_image_inference(event)
            return
        if event.event_type != events.EventType.LLM_INFERENCE_END:
            return
        
        if event.info.model_name not in self.models:
            self.models[event.info.model_name] = [0, 0, 0, 0, '']

        self.models[event.info.model_name][0] += 1
        self.models[event.info.model_name][1] += event.info.duration_s
        self.models[event.info.model_name][2] += event.info.input_tokens
        self.models[event.info.model_name][3] += event.info.output_tokens
        self.models[event.info.model_name][4] = event.info.provider_name




import axiomic.engine.node_impls.immediate as immediate
import axiomic.engine.node_impls.history_complete as history_complete
import axiomic.engine.node_impls.format_node as format_node
import axiomic.engine.node_impls.boolean_gate as boolean_gate
import axiomic.engine.node_impls.flat_rag_topk_node as flat_rag_topk_node
import axiomic.engine.node_impls.name_node as name_node 
import axiomic.engine.node_impls.txt_to_img_node as txt_to_img_node
import axiomic.engine.node_impls.builtin_binary_check_node as builtin_binary_check_node
import axiomic.engine.node_impls.builtin_unary_check_node as builtin_unary_check_node


import axiomic.protos as protos


NODES = [
    ('immediate_node', protos.axiomic.ImmediateNode, immediate.immediate),
    # ('coerce_json_node', protos.axiomic.CoerceJsonNode, coerce_json.coerce_json),
    ('history_complete_node', protos.axiomic.HistoryCompleteNode, history_complete.history_complete),
    ('format_node', protos.axiomic.FormatNode, format_node.format_node),
    # ('custom_check_node', protos.axiomic.CustomCheckNode, custom_check.custom_check),
    ('boolean_gate_node', protos.axiomic.BooleanGateNode, boolean_gate.boolean_gate),
    # ('coerce_boolean_node', protos.axiomic.CoerceBooleanNode, coerce_boolean.coerce_boolean),
    ('flat_rag_top_k_node', protos.axiomic.FlatRagTopKNode, flat_rag_topk_node.flat_rag_topk_node),
    ('name_node', protos.axiomic.NameNode, name_node.name_node),
    ('text_to_images_node', protos.axiomic.TextToImagesNode, txt_to_img_node.txt_to_img_node),
    ('builtin_binary_check_node', protos.axiomic.BuiltinBinaryCheckNode, builtin_binary_check_node.builtin_binary_check_node),
    ('builtin_unary_check_node', protos.axiomic.BuiltinUnaryCheckNode, builtin_unary_check_node.builtin_unary_check_node)
]


def field_from_proto(proto):
    type_ = str(type(proto))
    for node in NODES:
        if type_ == str(node[1]):
            return node[0]

def exec_hook_by_field_name(field_name):
    for node in NODES:
        if field_name == node[0]:
            return node[2]

    raise Exception(f'No node found for field name {field_name}')






import dataclasses
import numpy as np


@dataclasses.dataclass
class AEmbedding:
    '''
    This is typed container for an embedding vector which offers some convience functionality.

    This is a stateless class.

    TODO: WIP. Under construction.
    '''
    vector: np.ndarray

    def cosine_similarity(self, other):
        if len(self.vector) != len(other.vector):
            raise ValueError('Embeddings must be of the same length')

        dot_product = np.dot(self.vector, other.vector)
        norm_vector1 = np.linalg.norm(self.vector)
        norm_vector2 = np.linalg.norm(other.vector)
        similarity = dot_product / (norm_vector1 * norm_vector2)
        return similarity

    def normalize_l2(self):
        norm = np.linalg.norm(self.vector)
        if norm == 0:
            raise ValueError('Cannot normalize a zero vector')
        return AEmbedding(vector=self.vector / norm)




import jsonschema 
import axiomic.errors as errors
import axiomic.constants as constants
import axiomic.utils.wjson as wjson
import axiomic.protos as protos
import json


CHECKS = {}

def register_check(name):
    def wrapper(func):
        CHECKS[name] = func
        return func
    return wrapper


@register_check(protos.axiomic.BuiltinUnaryCheckNode.CHECK_JSON_SYNTAX)
def check_json_syntax(json_str: str) -> str:
    parse_ok = False
    try:
        jo = wjson.loads(json_str) 
        return constants.TRUE_VALUE
    except json.decoder.JSONDecodeError as e:
        pass

    return constants.FALSE_VALUE





import rich
import atexit
from rich import print
from rich.panel import Panel
from rich.text import Text
from rich.console import Group
from rich.traceback import install


'''
These explainers are separated from the rest because the main set of explainers
depends on the default configs.

These explainers have no deps within axiomic.
'''


ALMOST_THERE_WEAVE = Text("Almost there! There was a minor config uh oh.", style='bold yellow reverse', justify="center")
IF_NOT_YOUR_FAULT_REPORT = Text("If you believe this is an internal error, please file a github issue.", style='dim', justify="center")

AT_EXIT_EXPLAINERS = []

def exit_handler():
    if len(AT_EXIT_EXPLAINERS) == 0:
        return
    rich.print("\n")
    rich.print("\n")
    for explainer in AT_EXIT_EXPLAINERS:
        rich.print(explainer)
    rich.print()

atexit.register(exit_handler)


def explain_missing_path(name, file_path, examples_of_setting):
    title = "Path Not Found"
    l = [Text(e, style='dim', justify="left") for e in examples_of_setting]
    content = Group(
        Text(f"Please set or create {name}", style='dim', justify="center"),
        Text(f"\n{file_path}\n", style="red bold", justify="center"),
        Text(f"Consider one of the these...", style='dim', justify="left"),
        *l
    )

    explanation_panel = Panel(
        content,
        title=title,
        expand=False,
        border_style="bold yellow",
    )
    # AT_EXIT_EXPLAINERS.append(ALMOST_THERE_WEAVE)
    AT_EXIT_EXPLAINERS.append(explanation_panel)
    AT_EXIT_EXPLAINERS.append(IF_NOT_YOUR_FAULT_REPORT)
    raise ValueError(f'Missing {name}: {file_path}')





import axiomic.models.context as context
import axiomic.models.generic as generic


Generic = generic.Generic

# TogetherStableDiffusion_XL_Base_1_0 = context.Config(
#     image_provider_name='together_text',
#     image_model_name='stabilityai/stable-diffusion-xl-base-1.0'
# )
# 
# TogetherLlama3_8B_Chat = context.LLMConfig(
#     llm_provider_name='together_text',
#     llm_model_name='meta-llama/Llama-3-8b-chat-hf'
# )
# 
# TogetherLlama3_70B = context.LLMConfig(
#     llm_provider_name='together_text',
#     llm_model_name='meta-llama/Meta-Llama-3-70B'
# )
# 
# TogetherQuen1_5_35B_Chat = context.LLMConfig(
#     llm_provider_name='together_text',
#     llm_model_name='Qwen/Qwen1.5-32B-Chat'
# )

def TogetherChat(model):
    return context.LLMConfig(llm_provider_name='together_text', llm_model_name=model)

def TogetherEmbedding(model):
    # FIXME
    return context.Config(llm_provider_name='together_text', llm_model_name=model)

def TogetherImage(model):
    return context.Config(image_provider_name='together_img', image_model_name=model)

def TogetherCode(model):
    return context.Config(llm_provider_name='together_text', llm_model_name=model)

def TogetherLanguage(model):
    return context.Config(llm_provider_name='together_text', llm_model_name=model)

def TogetherModeration(model):
    return context.Config(llm_provider_name='together_text', llm_model_name=model)


class Together:

  class Text:

    class Austism:
      chronos_hermes_13b = TogetherChat("Austism/chronos-hermes-13b")

    class Nexusflow:
      nexusraven_v2_13b = TogetherChat("Nexusflow/NexusRaven-V2-13B")

    class NousResearch:
      nous_capybara_7b_v1p9 = TogetherChat("NousResearch/Nous-Capybara-7B-V1p9")
      nous_hermes_2_mistral_7b_dpo = TogetherChat("NousResearch/Nous-Hermes-2-Mistral-7B-DPO")
      nous_hermes_2_mixtral_8x7b_dpo = TogetherChat("NousResearch/Nous-Hermes-2-Mixtral-8x7B-DPO")
      nous_hermes_2_mixtral_8x7b_sft = TogetherChat("NousResearch/Nous-Hermes-2-Mixtral-8x7B-SFT")
      nous_hermes_2_yi_34b = TogetherChat("NousResearch/Nous-Hermes-2-Yi-34B")
      nous_hermes_llama2_13b = TogetherChat("NousResearch/Nous-Hermes-Llama2-13b")
      nous_hermes_llama_2_7b = TogetherChat("NousResearch/Nous-Hermes-llama-2-7b")

    class OpenOrca:
      mistral_7b_openorca = TogetherChat("Open-Orca/Mistral-7B-OpenOrca")

    class Phind:
      phind_codellama_34b_v2 = TogetherChat("Phind/Phind-CodeLlama-34B-v2")

    class Qwen:
      qwen1_5_0_5b = TogetherChat("Qwen/Qwen1.5-0.5B")
      qwen1_5_1_8b = TogetherChat("Qwen/Qwen1.5-1.8B")
      qwen1_5_110b_chat = TogetherChat("Qwen/Qwen1.5-110B-Chat")
      qwen1_5_14b = TogetherChat("Qwen/Qwen1.5-14B")
      qwen1_5_32b = TogetherChat("Qwen/Qwen1.5-32B")
      qwen1_5_4b = TogetherChat("Qwen/Qwen1.5-4B")
      qwen1_5_72b = TogetherChat("Qwen/Qwen1.5-72B")
      qwen1_5_7b = TogetherChat("Qwen/Qwen1.5-7B")

    class AllenAI:
      olmo_7b = TogetherChat("allenai/OLMo-7B")
      olmo_7b_twin_2t = TogetherChat("allenai/OLMo-7B-Twin-2T")

    class DeepseekAI:
      deepseek_llm_67b_chat = TogetherChat("deepseek-ai/deepseek-llm-67b-chat")

    class GarageBaind:
      platypus2_70b_instruct = TogetherChat("garage-bAInd/Platypus2-70B-instruct")

    class Google:
      gemma_2b = TogetherChat("google/gemma-2b")
      gemma_7b_it = TogetherChat("google/gemma-7b-it")

    class Microsoft:
      wizardlm_2_8x22b = TogetherChat("microsoft/WizardLM-2-8x22B")
      phi_2 = TogetherChat("microsoft/phi-2")

    class MistralAI:
      mistral_7b_instruct_v0_2 = TogetherChat("mistralai/Mistral-7B-Instruct-v0.2")
      mistral_7b_v0_1 = TogetherChat("mistralai/Mistral-7B-v0.1")
      mixtral_8x22b = TogetherChat("mistralai/Mixtral-8x22B")
      mixtral_8x7b_v0_1 = TogetherChat("mistralai/Mixtral-8x7B-v0.1")
      mistral_7b_instruct_v0_1 = TogetherLanguage("mistralai/Mistral-7B-Instruct-v0.1")
      mixtral_8x22b_instruct_v0_1 = TogetherLanguage("mistralai/Mixtral-8x22B-Instruct-v0.1")
      mixtral_8x7b_instruct_v0_1 = TogetherLanguage("mistralai/Mixtral-8x7B-Instruct-v0.1")

    class Openchat:
      openchat_3_5_1210 = TogetherChat("openchat/openchat-3.5-1210")

    class SnorkelAI:
      snorkel_mistral_pairrm_dpo = TogetherChat("snorkelai/Snorkel-Mistral-PairRM-DPO")

    class Teknium:
      openhermes_2_mistral_7b = TogetherChat("teknium/OpenHermes-2-Mistral-7B")
      openhermes_2p5_mistral_7b = TogetherChat("teknium/OpenHermes-2p5-Mistral-7B")

    class Together:
      gpt_jt_moderation_6b = TogetherChat("togethercomputer/GPT-JT-Moderation-6B")
      llama_2_7b_32k = TogetherChat("togethercomputer/LLaMA-2-7B-32K")
      redpajama_incite_7b_base = TogetherChat("togethercomputer/RedPajama-INCITE-7B-Base")
      stripedhyena_hessian_7b = TogetherChat("togethercomputer/StripedHyena-Hessian-7B")
      stripedhyena_nous_7b = TogetherChat("togethercomputer/StripedHyena-Nous-7B")

    class ZeroOneAI:
      yi_34b_chat = TogetherChat("zero-one-ai/Yi-34B-Chat")

    class Qwen:
      qwen1_5_0_5b_chat = TogetherLanguage("Qwen/Qwen1.5-0.5B-Chat")
      qwen1_5_1_8b_chat = TogetherLanguage("Qwen/Qwen1.5-1.8B-Chat")
      qwen1_5_14b_chat = TogetherLanguage("Qwen/Qwen1.5-14B-Chat")
      qwen1_5_32b_chat = TogetherLanguage("Qwen/Qwen1.5-32B-Chat")
      qwen1_5_4b_chat = TogetherLanguage("Qwen/Qwen1.5-4B-Chat")
      qwen1_5_72b_chat = TogetherLanguage("Qwen/Qwen1.5-72B-Chat")
      qwen1_5_7b_chat = TogetherLanguage("Qwen/Qwen1.5-7B-Chat")

    class Snowflake:
      snowflake_arctic_instruct = TogetherLanguage("Snowflake/snowflake-arctic-instruct")

    class AllenAI:
      olmo_7b_instruct = TogetherLanguage("allenai/OLMo-7B-Instruct")

    class CodeLlama:
      codellama_7b_python_hf = TogetherLanguage("codellama/CodeLlama-7b-Python-hf")

    class Cognitive:
      dolphin_2_5_mixtral_8x7b = TogetherLanguage("cognitivecomputations/dolphin-2.5-mixtral-8x7b")

    class DataBricks:
      dbrx_instruct = TogetherLanguage("databricks/dbrx-instruct")

    class DeepseekAi:
      deepseek_coder_33b_instruct = TogetherLanguage("deepseek-ai/deepseek-coder-33b-instruct")

    class Google:
      gemma_2b_it = TogetherLanguage("google/gemma-2b-it")
      gemma_7b = TogetherLanguage("google/gemma-7b")
    
    class Llama3:
      llama_3_8b_chat_hf = TogetherLanguage("meta-llama/Llama-3-8b-chat-hf")
      llama_3_8b_hf = TogetherChat("meta-llama/Llama-3-8b-hf")
      llama_3_70b_chat_hf = TogetherChat("meta-llama/Llama-3-70b-chat-hf")
      meta_llama_3_70b = TogetherLanguage("meta-llama/Meta-Llama-3-70B")

    class Llama2:
      llama_2_13b_hf = TogetherLanguage("meta-llama/Llama-2-13b-hf")
      llama_2_70b_hf = TogetherLanguage("meta-llama/Llama-2-70b-hf")
      llama_2_7b_chat_hf = TogetherLanguage("meta-llama/Llama-2-7b-chat-hf")
      llama_2_13b_chat_hf = TogetherChat("meta-llama/Llama-2-13b-chat-hf")
      llama_2_70b_chat_hf = TogetherChat("meta-llama/Llama-2-70b-chat-hf")
      llama_2_7b_hf = TogetherChat("meta-llama/Llama-2-7b-hf")
      llama_guard_7b = TogetherChat("Meta-Llama/Llama-Guard-7b")
      llamaguard_2_8b = TogetherChat("meta-llama/LlamaGuard-2-8b")

    class CodeLlama:
      codellama_13b_instruct_hf = TogetherCode("codellama/CodeLlama-13b-Instruct-hf")
      codellama_34b_instruct_hf = TogetherCode("codellama/CodeLlama-34b-Instruct-hf")
      codellama_70b_instruct_hf = TogetherCode("codellama/CodeLlama-70b-Instruct-hf")
      codellama_70b_python_hf = TogetherCode("codellama/CodeLlama-70b-Python-hf")
      codellama_7b_instruct_hf = TogetherCode("codellama/CodeLlama-7b-Instruct-hf")
      codellama_70b_hf = TogetherModeration("codellama/CodeLlama-70b-hf")
      codellama_13b_python_hf = TogetherChat("codellama/CodeLlama-13b-Python-hf")
      codellama_34b_python_hf = TogetherChat("codellama/CodeLlama-34b-Python-hf")

    # FIXME: Broken
    # class PromptHero:
    #  openjourney = TogetherLanguage("prompthero/openjourney")

    class Togethercomputer:
      llama_2_7b_32k_instruct = TogetherLanguage("togethercomputer/Llama-2-7B-32K-Instruct")
      redpajama_incite_7b_instruct = TogetherLanguage("togethercomputer/RedPajama-INCITE-7B-Instruct")
      redpajama_incite_chat_3b_v1 = TogetherLanguage("togethercomputer/RedPajama-INCITE-Chat-3B-v1")
      redpajama_incite_instruct_3b_v1 = TogetherLanguage("togethercomputer/RedPajama-INCITE-Instruct-3B-v1")
      evo_1_131k_base = TogetherLanguage("togethercomputer/evo-1-131k-base")
      evo_1_8k_base = TogetherLanguage("togethercomputer/evo-1-8k-base")

    class ZeroOneAI:
      yi_34b = TogetherLanguage("zero-one-ai/Yi-34B")
      yi_6b = TogetherLanguage("zero-one-ai/Yi-6B")

    class Wizardlm:
      wizardlm_13b_v1_2 = TogetherCode("WizardLM/WizardLM-13B-V1.2")

    class Lmsys:
      vicuna_13b_v1_5 = TogetherCode("lmsys/vicuna-13b-v1.5")

  class Embedding:

    class Baai:
      bge_base_en_v1_5 = TogetherEmbedding("BAAI/bge-base-en-v1.5")
      bge_large_en_v1_5 = TogetherEmbedding("BAAI/bge-large-en-v1.5")

    class WhereIsAI:
      uae_large_v1 = TogetherEmbedding("WhereIsAI/UAE-Large-V1")

    class SentenceTransformers:
      msmarco_bert_base_dot_v5 = TogetherEmbedding("sentence-transformers/msmarco-bert-base-dot-v5")

    class Together:
      m2_bert_80m_2k_retrieval = TogetherEmbedding("togethercomputer/m2-bert-80M-2k-retrieval")
      m2_bert_80m_32k_retrieval = TogetherEmbedding("togethercomputer/m2-bert-80M-32k-retrieval")
      m2_bert_80m_8k_retrieval = TogetherEmbedding("togethercomputer/m2-bert-80M-8k-retrieval")

    class Upstage:
      solar_10_7b_instruct_v1_0 = TogetherEmbedding("upstage/SOLAR-10.7B-Instruct-v1.0")

  class Image:

    class Sg161222:
      realisticvisionv3_0vae = TogetherChat("SG161222/Realistic_Vision_V3.0_VAE")

    class Wizardlm:
      wizardcoder_python_34b_v1_0 = TogetherImage("WizardLM/WizardCoder-Python-34B-V1.0")

    class StabilityAI:
      stable_diffusion_2_1 = TogetherImage("stabilityai/stable-diffusion-2-1")
      stable_diffusion_xl_base_1_0 = TogetherImage("stabilityai/stable-diffusion-xl-base-1.0")

    class Togethercomputer:
      redpajama_incite_base_3b_v1 = TogetherImage("togethercomputer/RedPajama-INCITE-Base-3B-v1")
      alpaca_7b = TogetherImage("togethercomputer/alpaca-7b")


  def bind():
    generic.bind(Generic.Text.Small, Together.Text.Llama3.llama_3_8b_chat_hf)
    generic.bind(Generic.Text.Medium, Together.Text.Qwen.qwen1_5_32b_chat)
    generic.bind(Generic.Text.Large, Together.Text.Llama3.llama_3_70b_chat_hf)
    generic.bind(Generic.Image.Medium, Together.Image.StabilityAI.stable_diffusion_2_1)






# Indicates the AI made a mistake, possibly retriable.
class ThoughtCrime(Exception):
    '''
    Thrown when the AI makes a mistake. Execution may be automatically retried, 
    if all retries fail, this will be raised to the caller.

    This will typically be thrown from unweaves, for example:

    .. code-block:: python

        try:
            name = name_axiomic.value()
        except axiomic.errors.ThoughtCrime:
            name = 'Unknown'
    '''
    pass


# Indicates the user built an incorrect weave, not retriable.
class RuntimeError(Exception):
    '''
    Thrown when there is an error in the way a weave was built.
    For example, if assinging an invalid name.
    '''
    def __init__(self, text=None, panel=None):
        self.panel = panel
        self.text = text

    def __str__(self):
        if self.panel:
            return str(self.panel)
        return self.text


# Indicates the user built an incorrect weave, not retriable.
class GraphError(Exception):
    '''
    Thrown when there is an error in the way a weave was built.
    For example, if assinging an invalid name.
    '''
    def __init__(self, text=None, panel=None):
        self.panel = panel
        self.text = text

    def __str__(self):
        if self.panel:
            return str(self.panel)
        return self.text


class ProviderErrorRetriable(Exception):
    '''
    Thrown when an error when calling an API which is re-triable.
    This includes, rate limit. 
    '''
    def __init__(self, message):
        self.message = message


class ProviderErrorNoRetry(Exception):
    '''
    An error with the API that should be retried.
    '''
    def __init__(self, message):
        self.message = message




# import axiomic.models.context as _context
# 
# 
# register_global_provider = _context.register_global_provider
# register_global_ref_provider = _context.register_global_ref_provider
# infer_context = _context.infer_context
# 
# get_provider_provider = _context.get_provider_provider
# print_inferred_context = _context.print_inferred_context
# 



import jsonschema 
import axiomic.errors as errors
import axiomic.constants as constants
import axiomic.utils.wjson as wjson
import axiomic.protos as protos
import json


CHECKS = {}


def register_check(name):
    def wrapper(func):
        CHECKS[name] = func
        return func
    return wrapper


@register_check(protos.axiomic.BuiltinBinaryCheckNode.CHECK_JSONSCHEMA)
def check_json_schema(json_str: str, json_schema: str) -> str:
    parse_ok = False
    try:
        jo = wjson.loads(json_str) 
        parse_ok = True
    except json.decoder.JSONDecodeError as e:
        pass

    if not parse_ok:
        raise errors.WeaveError('Ask to validate a JSON schema (did you forget to check json syntax first?) on a non-JSON string: ' + json_str)

    parse_ok = False
    try:
        schema = wjson.loads(json_schema)
        parse_ok = True
    except json.decoder.JSONDecodeError:
        pass

    if not parse_ok:
        raise errors.WeaveError('Schema is not valid JSON: ' + json_schema)

    try:
        jsonschema.validate(jo, schema)
        return constants.TRUE_VALUE
    except jsonschema.exceptions.ValidationError as e:
        print('/' * 80)
        print(e)
        print('\\' * 80)
    except jsonschema.exceptions.SchemaError as e:
        print('/' * 80)
        print(e)
        print('\\' * 80)
        raise errors.WeaveError('Schema is not a valid schema: ' + json_schema)

    return constants.FALSE_VALUE




from pydantic_settings import BaseSettings
from typing import Optional
from pydantic import field_validator, ValidationError, Field

import axiomic.configure.default_config_explainers as dc_explainers

import os


class FileRefStoreConfig(BaseSettings):
    data_path: Optional[str] = Field(default=None, pre=True)

    class Config:
        env_file = ".env"
        env_prefix = "WEAVE_FILE_"
    
    @field_validator('data_path')
    def set_abs_path_and_validate_existence(cls, v):
        if v is not None:
            abs_path = os.path.abspath(v)
            if not os.path.exists(abs_path):
                dc_explainers.explain_missing_path("file_ref_store.data_path", abs_path, 
                                                   [f'mkdir -p {abs_path}',
                                                    'export WEAVE_FILE_DATA_PATH=/path/to/your/data',
                                                    'default_config.file_ref_store.data_path = "/path/to/your/data"'])
                raise ValueError(f"The specified path does not exist: {abs_path}")
            return abs_path
        return None


class LLMConfig(BaseSettings):
    verbose: Optional[bool] = False

    class Config:
        env_file = ".env"
        env_prefix = "WEAVE_TEXT_"


class ConsoleConfig(BaseSettings):
    color_tracebacks: Optional[bool] = True
    help_on_exit: Optional[bool] = True
    enable: Optional[bool] = True

    class Config:
        env_file = ".env"
        env_prefix = "WEAVE_CONSOLE_"


file_ref_store = FileRefStoreConfig()
llm_config = LLMConfig()
console_config = ConsoleConfig()



from PIL import Image
from io import BytesIO
import requests
from rich.console import Console
import base64



class AImage:
    '''
    A container for a PIL Image with quality of life functions.
    
    Named to avoid conflict with the PIL Image class.

    TODO: WIP. Under construction.
    '''
    def __init__(self, img: Image):
        self.img = img

    def print(self, width=None):
        if width is None:
            console = Console()
            width, _ = console.size
        ansi = image_to_ansi(self.img, max_width=width)
        print(ansi)
        return self

    def __str__(self):
        return image_to_ansi(self.img, max_width=40)

    def save(self, path: str):
        self.img.save(path)


def image_from_base64(base64_str: str) -> Image:
    image_data = base64.b64decode(base64_str)
    image_data = BytesIO(image_data)
    image = Image.open(image_data)
    return AImage(image)


def image_from_url(url: str) -> Image:
    response = requests.get(url)
    image = Image.open(BytesIO(response.content))
    return AImage(image)


def image_to_ansi(img: Image, max_width=120, aspect_ratio=0.503):  # aspect_ratio adjusted for typical terminal character dimensions
    img = img.resize((max_width, int(max_width * img.height / img.width * aspect_ratio)))
    img = img.convert('RGB')
    
    ansi_image = ""
    for y in range(img.height):
        for x in range(img.width):
            pixel = img.getpixel((x, y))
            ansi_image += f"\x1b[48;2;{pixel[0]};{pixel[1]};{pixel[2]}m "  # Color the background
        ansi_image += '\x1b[0m\n'  # Reset the color at the end of each line
    
    return ansi_image



import axiomic.models.context as context
import axiomic.models.generic as generic

Generic = generic.Generic
Config = context.Config

class Anthropic:

    class Text:

        class Claude3:

            ''' Default Opus, you need to also `export ANTHROPIC_API_KEY=sk-...` '''
            Opus = Config(
                llm_provider_name='anthropic_text',
                llm_model_name='claude-3-opus-20240229',
                llm_temperature=0.5,
                llm_max_tokens=1024,
                _context_name='AnthropicOpus'
            )

            ''' Default Sonnet, you need to also `export ANTHROPIC_API_KEY=sk-...` '''
            Sonnet = Config(
                llm_provider_name='anthropic_text',
                llm_model_name='claude-3-sonnet-20240229',
                llm_temperature=0.5,
                llm_max_tokens=1024,
                _context_name='AnthropicSonnet'
            )

            ''' Default Haiku, you need to also `export ANTHROPIC_API_KEY=sk-...` '''
            Haiku = Config(
                llm_provider_name='anthropic_text',
                llm_model_name='claude-3-haiku-20240307',
                llm_temperature=0.5,
                llm_max_tokens=1024,
                _context_name='AnthropicHaiku'
            )
    
    def bind():
        generic.bind(Generic.Text.Small, Anthropic.Text.Claude3.Haiku)
        generic.bind(Generic.Text.Medium, Anthropic.Text.Claude3.Sonnet)
        generic.bind(Generic.Text.Large, Anthropic.Text.Claude3.Opus)




import threading

import axiomic.providers.provider as provider

import axiomic.configure.default_config as default_config

import rich
from rich.console import Console, Group
from rich.panel import Panel
from rich.text import Text
from rich import box
from rich.layout import Layout


from icecream import ic

'''
TODO: This file needs major cleanup.
'''

class ModelContext:
    def __init__(self, name=None):
        self.provider_provider = None
        self.llm_provider_name = None
        self.llm_model_name = None
        self.llm_temperature = None
        self.llm_max_tokens = None
        self.verbose_llm = None
        self._context_name = name
        self.image_provider_name = None
        self.image_model_name = None
        self.image_width = None
        self.image_height = None
        self.embedding_provider_name = None
        self.embedding_model_name = None
    
    def __str__(self):
        return str(self.__dict__)

    def bless_as_global(self):
        '''Does anything that needs to be done once for the global context.
        '''
        self.provider_provider = provider.ProviderProvider()

    def update_non_none(self, **kwargs):
        for k, v in kwargs.items():
            if v is not None:
                setattr(self, k, v)

    def update_all(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)
    
    def update_from(self, other, preserve_name=True):
        items = other.__dict__.copy()
        if preserve_name:
            del items['_context_name']
        self.update_non_none(**items)

    def __and__(self, other):
        neu = self.clone()
        neu.update_non_none(**other.__dict__)
        neu._context_name = f'{self._context_name} & {other._context_name}'
        return neu

    def clone(self):
        '''
        Clones the ModelContext.
        '''
        new_context = ModelContext()
        new_context.update_all(**self.__dict__)
        return new_context

    def __repr__(self):
        d = self.__dict__.copy()
        # filter out Nones
        d = {k: v for k, v in d.items() if v is not None}

        stuff = str(d)
        return f'ModelContext({stuff})'


class ContextStack:
    def __init__(self, root_config):
        self.stack = [root_config]

    def push_context(self, context):
        self.stack.append(context)
    
    def pop_context(self):
        if len(self.stack) == 1:
            raise Exception('Cannot pop the root context.')
        self.stack.pop()

    def get(self, name):
        stack = self.stack
        for c in reversed(self.stack):
            if hasattr(c, name) and getattr(c, name) is not None:
                return getattr(c, name)
        raise Exception(f'Could not resolve config name: {name}')

    def __getattr__(self, name):
        return self.get(name)


class ModelContextManager:
    _thread_local = threading.local()  # Thread-local storage

    def __init__(self, context: ModelContext):
        if isinstance(context, ContextStack):
            raise Exception('Should not be a context stack.')
        self.context = context
        if not hasattr(ModelContextManager._thread_local, 'current'):
            ModelContextManager._thread_local.current = ContextStack(GLOBAL_DEFAULT_CONTEXT)

    def __str__(self):
        return str(self.context.__dict__)

    def __enter__(self):
        if not hasattr(ModelContextManager._thread_local, 'current'):
            ModelContextManager._thread_local.current = ContextStack(GLOBAL_DEFAULT_CONTEXT)
        current = ModelContextManager._thread_local.current
        current.push_context(self.context)  # Set current context for this thread
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        current = ModelContextManager._thread_local.current
        current.pop_context()  # Clear current context for this thread

    def __and__(self, other):
        return ModelContextManager(self.context & other.context)

    def __repr__(self):
        return f'ModelContextManager(context={repr(self.context)}'


GLOBAL_DEFAULT_CONTEXT = ModelContext(name='Global Context')
GLOBAL_DEFAULT_CONTEXT.verbose_llm = default_config.llm_config.verbose
GLOBAL_DEFAULT_CONTEXT.bless_as_global()


def Config(_context_name=None, **kwargs):
    '''
    Context Manager which will set the values in the config.
    '''
    neu = ModelContext(name=_context_name)
    neu.update_all(**kwargs)
    return ModelContextManager(neu)


def LLMConfig(llm_provider_name: str = None, llm_model_name: str = None, llm_temperature: float = None, llm_max_tokens: int = None, _context_name=None):
    '''
    Creates an LLMSet context.
    '''
    if _context_name is None:
        _context_name = 'LLMConfig'
    return Config(llm_provider_name=llm_provider_name, llm_model_name=llm_model_name, llm_temperature=llm_temperature, llm_max_tokens=llm_max_tokens, _context_name=_context_name)


def infer_context():
    '''
    Attempts to determine the current ModelContext which applies to the caller.
    Returns None if it was unable to.
    '''
    possible_context = getattr(ModelContextManager._thread_local, 'current', None)
    if possible_context is not None:
        return possible_context
    return global_default_context()

def _context_to_panel(title, c):
    txt = ''
    for k, v in c.__dict__.items():
        if k == '_context_name':
            continue
        if k == 'provider_provider':
            continue
        if v is None:
            continue
        txt += f'{k}: {v}\n'
    txt = txt.strip()
    p = Panel(txt, title=title, style='dim')
    return p
    

def print_inferred_context():
    '''Pretty prints the inferred context.
    '''
    context = infer_context()
    
    if context == global_default_context():

        panel = _context_to_panel('GlobalContext', context) 
        # print('Global Default Context:', context)
    else:
        panels = []
        for i, c in enumerate(context.stack):
            context_name = c._context_name
            if context_name is None:
                context_name = f'(Anonymous Context)'
            panels.append(_context_to_panel(context_name, c))
        panel = Group(*panels)
        # print('Context Stack:', context.stack)
    context_stack = Panel(panel, title='Context Stack', style='dim', expand=True)
    console = Console()
    console.print(context_stack)
    


def global_default_context():
    return GLOBAL_DEFAULT_CONTEXT

def register_global_provider(provider):
    '''
    Registers a provider in the global context.
    '''
    return global_default_context().provider_provider.register_provider(provider)
    

def register_global_ref_provider(ref_provider):
    '''
    Registers a provider in the global context.
    '''
    global_default_context().provider_provider.ref_store_provider = ref_provider
    

def get_provider_provider():
    return infer_context().provider_provider
    



# Import List Types
from axiomic.data.loaders import load_strings, load_chat_text, load_jsons_unparsed, load_jsons_parsed, load_pydantics

# Load Chats
from axiomic.data.loaders import load_chat_text, load_chat_jsons_unparsed, load_chat_jsons_parsed, load_chat_pydantic


from axiomic.data.lists import TextList, ChatList





from typing import Any
import axiomic.protos as protos

import axiomic.engine.node_map as node_map
# import axiomic.context as context
import axiomic.models.context as wcontext

import axiomic.engine.exec_context as exec_context

import axiomic.graph.tree as tree
import axiomic.errors as errors
import axiomic.core_config as core_config
import hashlib
import axiomic.logalytics.sink as sink
import axiomic.logalytics.events as events


global_context = exec_context.ExecutionContext()


def md5sum_string(input_string):
    return hashlib.md5(input_string.encode()).hexdigest()


def hash_node(node: protos.axiomic.AxiomicNode) -> str:
    return md5sum_string(str(node))


class WeftContext:
    def __init__(self, graph: protos.axiomic.AxiomicGraph, exec_context: exec_context.ExecutionContext, config: core_config.Config):
        self.exec_context = exec_context
        self.retry_thought_crime_count = config.thought_crime_retry_count
        self._weft_cache = {}
        self.graph = tree.Graph(graph)

    def is_cached(self, node: any) -> bool:
        if hash_node(node) in self._weft_cache:
            return True
        if not core_config.global_config.enable_caching:
            return False
        ans = self.exec_context.is_cached(node)
        return ans

    def get_cached(self, node: any) -> str:
        if hash_node(node) in self._weft_cache:
            return self._weft_cache[hash_node(node)]

        if not core_config.global_config.enable_caching:
            raise Exception('Caching is disabled')

        return self.exec_context.get_cached(node)

    def cache_put(self, node: any, output: str):
        root_hash = hash_node(node)
        self._weft_cache[hash_node(node)] = output
        if not core_config.global_config.enable_caching:
            return
        self.exec_context.cache_put(node, output)

    def posion_cache(self, node: any):
        if hash_node(node) in self._weft_cache:
            del self._weft_cache[hash_node(node)]
        
        if not core_config.global_config.enable_caching:
            return
        self.exec_context.posion_cache(node)

    def resolve_node(self, node: any) -> str:
        if str(type(node)) == str(protos.axiomic.NodeRef):
            node = self.graph.ref_to_node(node)
        if not self.is_cached(node):
            raise Exception('Node not cached? Internal Error: ' + node.name)
        return self.get_cached(node)

    def get_provider(self, provider_name: str) -> Any:
        return wcontext.infer_context().provider_provider.get_provider(provider_name)
        

def print_graph(graph: protos.axiomic.AxiomicGraph, root: protos.axiomic.AxiomicNode):
    g = tree.Graph(graph)
    g.print_tree(root)

def eager_weft(graph: protos.axiomic.AxiomicGraph, node: protos.axiomic.AxiomicNode) -> str:
    weft = Weft(WeftContext(graph, global_context, config=core_config.global_config))
    global_context.flush_cache()
    return weft.exec_node(node)


class Weft:
    def __init__(self, context: WeftContext):
        self.context = context

    def _cache_node_output(self, node: protos.axiomic.AxiomicNode):
        if self.context.is_cached(node):
            return
        output = resolve_node(node, self.context)
        self.context.cache_put(node, output)

    def exec_node(self, node: protos.axiomic.AxiomicNode, blocking=True) -> str:
        if self.context.is_cached(node):
            return self.context.get_cached(node)

        tries_left = 1 + self.context.retry_thought_crime_count
        while tries_left > 0:
            try:
                root_hash = hash_node(node)
                self._flatten_and_cache_tree(node)
                # All nodes should be cached in the tree now. Let's get the root.

                if not self.context.is_cached(node):
                    raise Exception(f'Root node not cached??? {node.name} # {root_hash}')
                out = self.context.get_cached(node)
                return out
            except errors.ThoughtCrime as e:
                print('Weft: ThoughtCrime: ', e)
                tries_left -= 1
                if tries_left == 0:
                    raise e
                self._posion_cache(node)
                sink.event_sink(events.Event('Retry ThoughtCrime', events.EventType.THOUGHT_CRIME_RETRY, {'message': 'thought crime occured - retrying query.'}))

        raise Exception('Unreachable')

    def _posion_cache(self, node: protos.axiomic.AxiomicNode):
        # recursively removes this tree from the cache.
        all_nodes = tree.reversed_breadth_first_flatten(self.context.graph, node)
        for node in all_nodes:
            self.context.posion_cache(node)

    def _flatten_and_cache_tree(self, node: protos.axiomic.AxiomicNode):
        all_nodes = self.context.graph.flatten_node(node)
        for node in all_nodes:
            self._cache_node_output(node)

    def _exec_node(self, node: protos.axiomic.AxiomicNode, blocking=True) -> str:
        if not blocking:
            raise Exception('Non-blocking execution not supported yet')
        return resolve_node(node, self.context)


def get_context():
    c = context.Context()
    c.set_node_resolver(resolve_node)
    return c


def resolve_node(node: protos.axiomic.AxiomicNode, c: WeftContext) -> str:
    type_name = str(type(node))
    if type_name != str(protos.axiomic.AxiomicNode):
        raise Exception(type_name)

    field_name = node.WhichOneof('node_type')
    sub_node = getattr(node, field_name)
    output = node_map.exec_hook_by_field_name(field_name)(sub_node, c, node)
    return output


def run(node: any) -> str:
    return resolve_node(node, get_context())





import axiomic.protos as protos
import axiomic.engine as engine
import axiomic.engine.node_map as node_map
import axiomic.utils.naming as naming


COUNT = 0

def defualt_name(basename):
    global COUNT
    COUNT += 1
    basename = basename.replace('_node', '')
    return f'_{basename}{COUNT}'


class GraphBuilder:
    
    def __init__(self):
        self.nodes = []
        self.node_map = {}

    def get_graph(self):
        wg = protos.axiomic.AxiomicGraph()
        for node in self.nodes:
            wg.nodes.append(node)
        return wg

    def add_node(self, node: protos.axiomic.AxiomicNode):
        if node.name in self.node_map:
            return
        self.nodes.append(node)
        self.node_map[node.name] = node

    def upbox_node(self, node, name=None):
        if name is not None:
            name = naming.Name(name)
        field_name = node_map.field_from_proto(node)
        w = protos.axiomic.AxiomicNode()
        getattr(w, field_name).CopyFrom(node)
        if name:
            w.name = str(name)
        else:
            w.name = defualt_name(field_name)
        self.add_node(w)
        return w

    def autobox_as_node(self, anything, name=None):
        if isinstance(anything, str):
            return self.upbox_node(protos.axiomic.ImmediateNode(value=anything), name=name)

        if 'axiomic' in str(type(anything)):
            return anything
        
        anything_type = type(anything)
        raise Exception(f'Cannot autobox {anything_type}: {anything}')
    
    def merge_from(self, other):
        for node in other.nodes:
            if node.name not in self.node_map:
                self.add_node(node)

    def print_graph(self, root: protos.axiomic.AxiomicNode):
        return self.get_graph().print_tree(root)


def weave(gb: GraphBuilder, anything, name=None):
    node = gb.autobox_as_node(anything, name=name)
    gb.add_node(node)
    return node


def infer_image(gb: GraphBuilder, image_prompt: any, params: protos.axiomic.TextToImagesParams, name=None):
    prompt_node = gb.autobox_as_node(image_prompt)
    node = protos.axiomic.TextToImagesNode()
    node.image_prompt.name = prompt_node.name
    node.params.CopyFrom(params)
    return gb.upbox_node(node, name=name)


def complete(gb: GraphBuilder, context_window: any, system_prompt: any, complete_params: protos.axiomic.CompleteParams, history_pairs: any = [], name=None):
    context_window = gb.autobox_as_node(context_window)
    system_prompt = gb.autobox_as_node(system_prompt)

    flat_complete_node = protos.axiomic.HistoryCompleteNode()
    flat_complete_node.user_message.name = context_window.name
    flat_complete_node.system_prompt.name = system_prompt.name

    for user_msg, agent_msg in history_pairs:
        user_message = gb.autobox_as_node(user_msg)
        agent_message = gb.autobox_as_node(agent_msg)
        history_pair = protos.axiomic.ChatHistoryPair()
        history_pair.user_message.name = user_message.name
        history_pair.assistant_message.name = agent_message.name
        flat_complete_node.chat_history.append(history_pair)

    flat_complete_node.complete_params.CopyFrom(complete_params)
    n = gb.upbox_node(flat_complete_node, name=name)
    return n


def format(gb: GraphBuilder, template: any, **inputs):
    template = gb.autobox_as_node(template)
    format_node = protos.axiomic.FormatNode()
    format_node.template.name = template.name
    for k, v in inputs.items():
        boxed_v = gb.autobox_as_node(v)
        format_input = protos.axiomic.FormatInput(name=k)
        format_input.value.name = boxed_v.name
        format_node.inputs.append(format_input)
    return gb.upbox_node(format_node)


def coerce_json(gb: GraphBuilder, value: any):
    value = gb.autobox_as_node(value)
    coerce_json_node = protos.axiomic.CoerceJsonNode(json_ish=value)
    return gb.upbox_node(coerce_json_node)


def name_node(gb: GraphBuilder, value: any, name):
    value = gb.autobox_as_node(value)
    name_node = protos.axiomic.NameNode()
    name_node.input.name = value.name
    return gb.upbox_node(name_node, name=name)


def check_json_schema(gb: GraphBuilder, json_object: any, schema: any):
    return builtin_binary_check(gb, json_object, schema, protos.axiomic.BuiltinBinaryCheckNode.CHECK_JSONSCHEMA)


def builtin_binary_check(gb: GraphBuilder, input_main: any, input_secondary: any, binary_check_name: any):
    main = gb.autobox_as_node(input_main)
    secondary = gb.autobox_as_node(input_secondary)
    check_node = protos.axiomic.BuiltinBinaryCheckNode()
    check_node.input_main.name = main.name
    check_node.input_secondary.name = secondary.name
    check_node.binary_check_name = binary_check_name
    return gb.upbox_node(check_node)


def check_json_syntax(gb: GraphBuilder, json_object: any):
    return builtin_unary_check(gb, json_object, protos.axiomic.BuiltinUnaryCheckNode.CHECK_JSON_SYNTAX)


def core_unary_check(gb: GraphBuilder, input_value: any, unary_check_name: any):
    input_value = gb.autobox_as_node(input_value)
    check_node = protos.axiomic.BuiltinUnaryCheckNode()
    check_node.input.name = input_value.name
    check_node.unary_check_name = unary_check_name
    return gb.upbox_node(check_node)


def flat_rag_topk_query(gb: GraphBuilder, query: any, separator: any, k: int, rag_provider: str):
    query = gb.autobox_as_node(query)
    seperator = gb.autobox_as_node(separator)
    rag_node = protos.axiomic.FlatRagTopKNode()

    rag_node.query.name = query.name
    rag_node.separator.name = seperator.name
    rag_node.k = k
    rag_node.rag_provider = rag_provider
    return gb.upbox_node(rag_node)


def coerce_boolean(gb: GraphBuilder, value: any):
    value = gb.autobox_as_node(value)
    coerce_boolean_node = protos.axiomic.CoerceBooleanNode(boolean_ish=value)
    # FXIME - make this configurable
    coerce_boolean_node.uncertain_mode = protos.axiomic.CoerceBooleanNode.UNCERTAIN_IS_THOUGHT_CRIME
    return gb.upbox_node(coerce_boolean_node)


def boolean_gate(gb: GraphBuilder, boolean: any, gating: any):
    boolean = gb.autobox_as_node(boolean)
    gating = gb.autobox_as_node(gating)

    boolean_gate_node = protos.axiomic.BooleanGateNode()

    boolean_gate_node.value.name = boolean.name
    boolean_gate_node.gating.name = gating.name

    return gb.upbox_node(boolean_gate_node)




import axiomic.errors as errors
import axiomic

import rich

import atexit
import os
import json


'''
def bad_name_explainer(name):
    """
    Given a bad name, return an explanation of why it is bad.
    """
'''

from rich import print
from rich.panel import Panel
from rich.text import Text
from rich.console import Group
import axiomic.errors as errors
import sys

import axiomic.configure.default_config as default_config

from rich.traceback import install


if default_config.console_config.enable:
    rich.print("[dim](axiomic) weave in helpful console mode (export AXIOMIC_CONSOLE_ENABLE=False to disable).[/dim]")

if default_config.console_config.enable and default_config.console_config.color_tracebacks:
    install(show_locals=False) #, suppress=[weave, json])

OLD_SYS_EXCEPT_HOOK = sys.excepthook

def thought_crime_exception_handler(exc_type, exc_value, exc_traceback):
    if exc_type is errors.ThoughtCrime:
        dying_to_thoughtcrime(exc_value)
    OLD_SYS_EXCEPT_HOOK(exc_type, exc_value, exc_traceback)


RAN_ONCE = {}

def only_once(f):
    def wrapped(*args, **kwargs):
        if f in RAN_ONCE:
            return
        RAN_ONCE[f] = True
        return f(*args, **kwargs)
    return wrapped


# Set the custom exception handler
sys.excepthook = thought_crime_exception_handler

ALMOST_THERE_WEAVE = Text("Axiomic: Almost there! A minor issue with the weave was found.", style='bold yellow reverse', justify="center")

THOUGHT_CRIME_BANNER = Text("Axiomic: Exit due to uncaught Thoguht Crime.", style='bold red reverse', justify="center")

IF_NOT_YOUR_FAULT_REPORT = Text("If you believe this is an internal error, please file a github issue.", style='dim', justify="center")

FOR_GRUMPY = Text("""export AXIOMIC_CONSOLE_COLOR_TRACEBACKS=False; export AXIOMIC_CONSOLE_HELP_ON_EXIT=False; # if you'd prefer a simpler experience.""", style='dim', justify="left")

AT_EXIT_EXPLAINERS = []

def exit_handler():
    if len(AT_EXIT_EXPLAINERS) == 0:
        return
    rich.print(FOR_GRUMPY)
    rich.print()
    for explainer in AT_EXIT_EXPLAINERS:
        rich.print(explainer)
    rich.print()

if default_config.console_config.enable and default_config.console_config.help_on_exit:
    atexit.register(exit_handler)


def explain_missing_file(file_path):
    title = "File Needs Creation"
    content = Group(
        Text("Please create this file.", style='dim', justify="center"),
        Text(f"\n{file_path}\n", style="red bold", justify="center"),
    )

    explanation_panel = Panel(
        content,
        title=title,
        expand=False,
        border_style="bold yellow",
    )
    AT_EXIT_EXPLAINERS.append(ALMOST_THERE_WEAVE)
    AT_EXIT_EXPLAINERS.append(explanation_panel)
    AT_EXIT_EXPLAINERS.append(IF_NOT_YOUR_FAULT_REPORT)
    raise errors.GraphError(text=f'Missing File: {file_path}')

def explain_wrapped_name(bad_name):
    title = "Invalid Name"

    content = Group(
        Text("Name incorrectly wrapped in '{' '}'   ", style='dim', justify="center"),
        Text(f"\nName: {bad_name}\n", style="red bold", justify="center"),
        Text("Try removing the { } around the name.", style='dim', justify="center"),
        Text("This name is not being used in a subsitution.", style='dim', justify="center"),
    )

    explanation_panel = Panel(
        content,
        title=title,
        # expand=False,
        border_style="bold yellow",
        # justify="center",
    )

    g = Group(ALMOST_THERE_WEAVE, explanation_panel, IF_NOT_YOUR_FAULT_REPORT)
    # AT_EXIT_EXPLAINERS.append(ALMOST_THERE_WEAVE)
    # AT_EXIT_EXPLAINERS.append(explanation_panel)
    # AT_EXIT_EXPLAINERS.append(IF_NOT_YOUR_FAULT_REPORT)
    AT_EXIT_EXPLAINERS.append(g)
    raise errors.GraphError(text=f'Bad Name: {bad_name}')

def explain_bad_name_error(bad_name):
    title = "Invalid Name"

    content = Group(
        Text("Name does not match [BUCKET:][DIR:]BASENAME   ", style='dim', justify="center"),
        Text(f"\nName: {bad_name}\n", style="red bold", justify="center"),
        Text("(optional) BUCKET    ~ [a-zA-Z][a-zA-Z0-9_.-]+", style='dim', justify="center"),
        Text("(optional) DIR       ~ [a-zA-Z][a-zA-Z0-9_.-]+", style='dim', justify="center"),
        Text("           BASENAME  ~ [a-zA-Z][a-zA-Z0-9_.-]+", style='dim', justify="center"),
        Text("Example: code_gen_prompt", style='dim', justify="center"),
        Text("Example: prompts:chat_prompts:friendly_hello", style='dim', justify="center"),
    )

    explanation_panel = Panel(
        content,
        title=title,
        # expand=False,
        border_style="bold yellow",
        # justify="center",
    )

    g = Group(ALMOST_THERE_WEAVE, explanation_panel, IF_NOT_YOUR_FAULT_REPORT)
    # AT_EXIT_EXPLAINERS.append(ALMOST_THERE_WEAVE)
    # AT_EXIT_EXPLAINERS.append(explanation_panel)
    # AT_EXIT_EXPLAINERS.append(IF_NOT_YOUR_FAULT_REPORT)
    AT_EXIT_EXPLAINERS.append(g)
    raise errors.GraphError(text=f'Bad Name: {bad_name}')


def missing_file_ref(file_ref, file_path):
    title = "File Ref Not Found"

    content = Group(
        Text("Looking for this ref", style='dim', justify="center"),
        Text(f"\n{file_ref}\n", style="red bold", justify="center"),
        Text("Leads to this fle", style='dim', justify="center"),
        Text(f"\n{file_path}\n", style="red bold", justify="center"),
        Text("Consider creating the file or changing the ref name. ", style='dim', justify="left"),
        Text(f"mkdir -p {os.path.dirname(file_path)} ", style='dim', justify="left"),
        Text(f"touch {file_path} ", style='dim', justify="left"),
    )

    explanation_panel = Panel(
        content,
        title=title,
        # expand=False,
        border_style="bold yellow",
        # justify="center",
    )

    g = Group(ALMOST_THERE_WEAVE, explanation_panel, IF_NOT_YOUR_FAULT_REPORT)
    # AT_EXIT_EXPLAINERS.append(ALMOST_THERE_WEAVE)
    # AT_EXIT_EXPLAINERS.append(explanation_panel)
    # AT_EXIT_EXPLAINERS.append(IF_NOT_YOUR_FAULT_REPORT)
    AT_EXIT_EXPLAINERS.append(g)
    raise errors.GraphError(text=f'Missing File Reference: {file_ref}')


@only_once
def set_file_store_path():
    title = "Set File Store Path"

    content = Group(
        Text("Please set the file store path, for example...", style='dim', justify="center"),

        Text('mkdir -p ./my_weave_data', style='dim', justify="left"),
        Text('THEN', style='dim', justify="center"),
        Text("import axiomic.configure.quick as quick", style='dim', justify="left"),
        Text('quick.set_relative_weave_data_root("./weave_data")', style='dim', justify="left"),
        Text('OR', style='dim', justify="center"),
        Text("export WEAVE_FILE_DATA_PATH=./my_weave_data", style='dim', justify="left"),
    )

    explanation_panel = Panel(
        content,
        title=title,
        # expand=False,
        border_style="bold yellow",
        # justify="center",
    )

    g = Group(ALMOST_THERE_WEAVE, explanation_panel, IF_NOT_YOUR_FAULT_REPORT)
    # AT_EXIT_EXPLAINERS.append(ALMOST_THERE_WEAVE)
    # AT_EXIT_EXPLAINERS.append(explanation_panel)
    # AT_EXIT_EXPLAINERS.append(IF_NOT_YOUR_FAULT_REPORT)
    AT_EXIT_EXPLAINERS.append(g)
    raise errors.GraphError(text=f'File Store Path Not Set')

def dying_to_thoughtcrime(tc):
    title = "axiomic.errors.ThoughtCrime"

    content = Group(
        Text(str(tc), style='dim', justify="left"),
    )

    explanation_panel = Panel(
        content,
        title=title,
        border_style="bold red",
    )

    g = Group(THOUGHT_CRIME_BANNER, explanation_panel, IF_NOT_YOUR_FAULT_REPORT)
    AT_EXIT_EXPLAINERS.append(g)

# Example usage
#bad_name = "invalid_name!"
#explain_bad_name_error(bad_name)

# explain_missing_file("/Users/victor/repos/weave/weave/logalytics/events.py")





import axiomic.protos as protos

import json
import re


'''
message FormatInput {
    // The name of this input, for example, "my_name"
    string name = 1;

    // The value to format into the template, for example, "Victor"
    AxiomicNode value = 2;
}

message FormatNode {
    // A template, e.g. "Hello, my name is {my_name}."
    AxiomicNode template = 1;

    // The values to fill in the template, e.g. "name" -> "Victor".
    repeated FormatInput inputs = 2;
}
'''

KEY_REGEX = re.compile(r'{([a-zA-Z0-9_]+)}')

def detect_keys(text):
    return KEY_REGEX.findall(text)


def format_node(fn: protos.axiomic.FormatNode, context, weave_node: protos.axiomic.AxiomicNode) -> str:
    template = context.resolve_node(fn.template)
    inputs = fn.inputs
    # Replace the template with the values
    input_map = {}
    for i in inputs:
        name = i.name
        value = context.resolve_node(i.value)
        input_map[name] = value
    
    all_keys = detect_keys(template)
    missing_keys = set(all_keys) - set(input_map.keys())
    place_holders = {k: f'{{{k}}}' for k in missing_keys}
    input_map.update(place_holders)
    try:
        out = template.format(**input_map)
    except KeyError:
        raise ValueError(f'Failed to format template: {template} with inputs: {input_map}')
    return out





import axiomic.protos as protos
import axiomic.errors as errors
import axiomic.engine.node_impls.builtin_binary_checks as builtin_binary_checks 


def builtin_binary_check_node(bucn: protos.axiomic.BuiltinBinaryCheckNode, context, weave_node: protos.axiomic.AxiomicNode) -> str:
    incoming_str = context.resolve_node(bucn.input_main)
    secondary_str  = context.resolve_node(bucn.input_secondary)
    binary_check_name = bucn.binary_check_name

    if binary_check_name not in builtin_binary_checks.CHECKS:
        raise errors.WeaveError(f'Unknown builtin binary check: {binary_check_name}')

    answer = builtin_binary_checks.CHECKS[binary_check_name](incoming_str, secondary_str)
    return answer



import atexit

import axiomic.logalytics.events as events


import axiomic.logalytics.console_log as console_log 
import axiomic.logalytics.stats_counter as stats_counter 

import axiomic.configure.default_config as default_config

STATS_COUNTER = stats_counter.StatsCounter()

def exit_handler():
    if STATS_COUNTER is not None:
        STATS_COUNTER.print_stats()

if default_config.console_config.enable:
    atexit.register(exit_handler)

HANDLERS = []

if default_config.console_config.enable:
    HANDLERS.append(console_log.ConsoleLog())
    HANDLERS.append(STATS_COUNTER)


def event_sink(event: events.Event):
    for handler in HANDLERS:
        handler.log_event(event)

def discover_config(message):
    event = events.Event('config', events.EventType.CONFIG_DISCOVER, {'message': message})
    event_sink(event)
    



import collections
from typing import List
from typing import List, Tuple, Union

import pydantic

from rich import print
from rich.panel import Panel
from rich.columns import Columns
from rich.syntax import Syntax
from rich.json import JSON
from rich.console import Console
from rich.layout import Layout
from rich.console import Group
from rich.text import Text

console = Console()

ChatPair = collections.namedtuple('ChatPair', ['user', 'agent'])

class TextList:
    
    def __init__(self, strings: any = [], name: str = None):
        if isinstance(strings, TextList):
            self.strings = strings.strings
        elif isinstance(strings, list):
            self.strings = strings
        self.name = name
        self.strigns = strings

    def __len__(self):
        return len(self.strings)

    def __getitem__(self, index):
        return self.strings[index]

    def __iter__(self):
        return iter(self.strings)

    def print(self):
        _print_list(self)



TextListType = Union[TextList, List[any]]


class ChatList:
    def __init__(self, pairs: any = None, name: str = None):
        self.name = name
        if pairs is None:
            self.pairs = []
        elif isinstance(pairs, ChatList):
            self.pairs = pairs.pairs
        else:
            # TODO add more type checking before assuming it is a list of tuples
            self.pairs = [ChatPair(user, agent) for user, agent in pairs]

    def __len__(self):
        return len(self.pairs)

    def __getitem__(self, index):
        return self.pairs[index]

    def __iter__(self):
        return iter(self.pairs)

    def append(self, user, agent):
        self.pairs.append(ChatPair(user, agent))

    def as_list(self):
        return [(pair.user, pair.agent) for pair in self.pairs]

    def extend(self, pairs):
        self.pairs.extend(ChatList(pairs).pairs)

    def copy(self):
        return ChatList(list(self.pairs))

    def make_eval_varients(self):
        varients = []
        for i in range(len(self.pairs)):
            cur = self.pairs[i]
            all_others = self.pairs[:i] + self.pairs[i+1:]
            varients.append((cur, ChatList(all_others, self.name)))
        return varients

    def print(self):
        _print_chat(self)

    def format(self, user_title="user", agent_title="agent", major_delim="\n============\n", minor_delim="\n---------\n", most_recent_user=None):
        txt = major_delim.join([f"{user_title}: {pair.user}{minor_delim}{agent_title}: {pair.agent}" for pair in self.pairs])

        if most_recent_user is not None:
            txt += f"{major_delim}{user_title}: {most_recent_user}"

        return txt


def _convert_for_print(thing):
    if isinstance(thing, str):
        return thing, 'string' #Syntax(thing, "json")
    if isinstance(thing, pydantic.BaseModel):
        o = JSON(thing.json())
        return o, f"pydantic<{type(thing).__name__}>"
    return JSON(thing), 'json'


def _print_chat(chat_list: ChatList):
    # Create a list of panel groups
    def create(c, idx):
        u, u_footer = _convert_for_print(c.user)
        a, a_footer = _convert_for_print(c.agent)
        panel_u = Panel(u, title="User" + str(idx), width=int(console.width // 2) - 2, subtitle=u_footer)
        panel_a = Panel(a, title="Agent" + str(idx), width=int(console.width // 2) - 2, subtitle=a_footer)
        return Columns([
            panel_u, panel_a
        ], expand=True )
    panel_groups = [ create(c, i) for i, c in enumerate(chat_list) ]
    # Display each group of panels
    for group in panel_groups:
        print(group)


def _print_list(wlist: TextList):
    # Create a list of panel groups
    def create(c, idx):
            item, footer = _convert_for_print(c)
            return Panel(item, title="Item " + str(idx), width=int(console.width) - 2, subtitle=footer)
    panel_groups = [ create(c, i) for i, c in enumerate(wlist) ]
    # Display each group of panels
    for group in panel_groups:
        print(group)


ChatListType = Union[ChatList, List[Tuple[any, any]]]

if __name__ == '__main__':
    print(str(ChatList([('a', 'b'), ('c', 'd')]).pairs))




import axiomic.protos as protos

import axiomic.errors as errors
import axiomic.constants as constants

VALUES_FOR_TRUE = [
    constants.TRUE_VALUE
]

def boolean_gate(bgn: protos.axiomic.BooleanGateNode, context, weave_node: protos.axiomic.AxiomicNode) -> str:
    value = context.resolve_node(bgn.value)
    gating = context.resolve_node(bgn.gating)

    if gating in VALUES_FOR_TRUE:
        return value

    raise errors.ThoughtCrime(f'Boolean Gated Failed: Expected Gating to be "True". Gate=<{gating}> Blocking Value=<{value}>')



import axiomic.core_config as core_config

import sqlite3

DB_PATH = '/tmp/axiomic_cache.db'

def should_cache():
    return core_config.global_config.enable_caching

class ExecutionContext:
    def __init__(self, disk_cache=True):
        if not core_config.global_config.enable_file_cache:
            disk_cache = False
        self._mem_cache = {}
        self._mem_cache_enabled = True
        self._disk_cache_enabled = disk_cache
        if self._disk_cache_enabled:
            self._load_disk_cache()

    def __del__(self):
        self.flush_cache()

    def _load_disk_cache(self):
        self._mem_cache.update(load_cache(DB_PATH))

    def set_mem_cache_enable(self, enable: bool):
        self._mem_cache_enabled = enable

    def flush_cache(self):
        if self._disk_cache_enabled:
            save_cache(DB_PATH, self._mem_cache)

    def is_cached(self, node: any) -> bool:
        if not self._mem_cache_enabled:
            return False
        nodestr = str(node)
        return nodestr in self._mem_cache

    def get_cached(self, node: any) -> str:
        nodestr = str(node)
        return self._mem_cache[nodestr]

    def cache_put(self, node: any, value: str):
        nodestr = str(node)
        self._mem_cache[nodestr] = value

    def posion_cache(self, node: any):
        nodestr = str(node)
        if nodestr in self._mem_cache:
            del self._mem_cache[nodestr]


def load_cache(db_path):
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()
    cursor.execute("CREATE TABLE IF NOT EXISTS cache (key TEXT PRIMARY KEY, value TEXT)")
    cursor.execute("SELECT key, value FROM cache")
    cache = dict(cursor.fetchall())
    print('exec_context.load_cache: loaded cache of ', len(cache))
    conn.close()
    return cache

def save_cache(db_path, cache):
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()
    cursor.execute("CREATE TABLE IF NOT EXISTS cache (key TEXT PRIMARY KEY, value TEXT)")
    cursor.executemany("REPLACE INTO cache (key, value) VALUES (?, ?)", cache.items())
    conn.commit()
    conn.close()




import os

import axiomic.configure.default_config as default_config
import axiomic.models.discover as discover
import axiomic.models.context as context


# These are exported to users

from axiomic.models.generic import Generic, bind

from axiomic.models.models_together import Together
from axiomic.models.models_openai import OpenAI
from axiomic.models.models_anthropic import Anthropic


discover.discover_providers()

Config = context.Config
'''
Create a new config, for exmaple,

    .. code-block:: python

        import axiomic.models
        GPT3_5 = axiomic.models.Config(
            llm_provider_name='openai_text',
            llm_model_name='gpt-3.5-turbo',
            llm_temperature=0.5,
            llm_max_tokens=1024,
            _context_name='OpenAiGpt3_5'
        )

        with GP3_5:
            axiomic.infer('What is the meaning of life?').print()

'''


VerboseText = context.Config(verbose_llm=True, _context_name='VerboseText')
'''
Makes text inference verbose. For example;

    .. code-block:: python
        import axiomic.models
        with axiomic.models.VerboseText:
            response = axiomic.infer('How much wood would a woodchuck chuck?', system_prompt='Respond in only 1 sentence.').value()

'''

MaxTokens32 = context.LLMConfig(llm_max_tokens=32, _context_name='MaxTokens32')
''' Sets Max Output Tokens to 128 '''

MaxTokens64 = context.LLMConfig(llm_max_tokens=64, _context_name='MaxTokens64')
''' Sets Max Output Tokens to 128 '''

MaxTokens128 = context.LLMConfig(llm_max_tokens=128, _context_name='MaxTokens128')
''' Sets Max Output Tokens to 128 '''

MaxTokens256 = context.LLMConfig(llm_max_tokens=256, _context_name='MaxTokens256')
"""Sets Max Output Tokens to 256"""

MaxTokens512 = context.LLMConfig(llm_max_tokens=512, _context_name='MaxTokens512')
"""Sets Max Output Tokens to 512"""

MaxTokens1024 = context.LLMConfig(llm_max_tokens=1024, _context_name='MaxTokens1024')
"""Sets Max Output Tokens to 1024"""

MaxTokens2048 = context.LLMConfig(llm_max_tokens=2048, _context_name='MaxTokens2048')
"""Sets Max Output Tokens to 2048"""

MaxTokens4096 = context.LLMConfig(llm_max_tokens=4096, _context_name='MaxTokens4096')
"""Sets Max Output Tokens to 4096"""

MaxTokens8192 = context.LLMConfig(llm_max_tokens=8192, _context_name='MaxTokens8192')
"""Sets Max Output Tokens to 8192"""


Temperature0_0 = context.LLMConfig(llm_temperature=0.0, _context_name='Temperature0_0')
''' Sets LLM temperature to 0.0 '''

Temperature0_1 = context.LLMConfig(llm_temperature=0.1, _context_name='Temperature0_1')
''' Sets LLM temperature to 0.1 '''

Temperature0_2 = context.LLMConfig(llm_temperature=0.2, _context_name='Temperature0_2')
''' Sets LLM temperature to 0.2 '''

Temperature0_3 = context.LLMConfig(llm_temperature=0.3, _context_name='Temperature0_3')
''' Sets LLM temperature to 0.3 '''

Temperature0_4 = context.LLMConfig(llm_temperature=0.4, _context_name='Temperature0_4')
''' Sets LLM temperature to 0.4 '''

Temperature0_5 = context.LLMConfig(llm_temperature=0.5, _context_name='Temperature0_5')
''' Sets LLM temperature to 0.5 '''

Temperature0_6 = context.LLMConfig(llm_temperature=0.6, _context_name='Temperature0_6')
''' Sets LLM temperature to 0.6 '''

Temperature0_7 = context.LLMConfig(llm_temperature=0.7, _context_name='Temperature0_7')
''' Sets LLM temperature to 0.7 '''

Temperature0_8 = context.LLMConfig(llm_temperature=0.8, _context_name='Temperature0_8')
''' Sets LLM temperature to 0.8 '''

Temperature0_9 = context.LLMConfig(llm_temperature=0.9, _context_name='Temperature0_9')
''' Sets LLM temperature to 0.9 '''

Temperature1_0 = context.LLMConfig(llm_temperature=1.0, _context_name='Temperature1_0')
''' Sets LLM temperature to 1.0 '''


def global_default(model_config):
    '''
    Updates the global default context. For example,

    .. code-block:: python

        import axiomic.models as models
        models.global_default(models.Generic.Text.Large)
        models.global_default(models.Temperature0_7)
        models.global_default(models.MaxTokens2048)

    '''
    context.GLOBAL_DEFAULT_CONTEXT.update_from(model_config.context)
    
def infer_context():
    '''
    Returns the current context.

    For example,

    .. code-block:: python

        import axiomic.models as models
        with models.SmallText & models.MaxTokens256 & models.Temperature0_5:
            print(models.infer_context())

    '''
    return context.infer_context()

def print_context():
    '''
    Prints the current context to the console.

    For example,
    
    .. code-block:: python

        import axiomic.configure.quick as wc
        with wc.SmallText & wc.MaxTokens256 & wc.Temperature0_5:
            wc.print_context()

    '''
    context.print_inferred_context()


def get_weave_data_root():
    if default_config.file_ref_store.data_path is None:
        return ''
    return default_config.file_ref_store.data_path


def set_relative_data_root(relative_path: str):
    """
    Sets the relative path to the weave data root directory.

    Args:
        relative_path (str): The relative path to the weave data root directory.
    """
    user_filename = discover.discover_user_calling_file()
    user_dirname = os.path.dirname(user_filename)
    weave_path = os.path.join(user_dirname, relative_path)
    default_config.file_ref_store.data_path = weave_path


def resolve_axiomic_variables(text):
    """
    Resolves weave variables in the text.

    Args:
        text (str): The text to resolve variables in.

    Returns:
        str: The text with variables resolved.
    """
    replaces = {
        '$WEAVE_DATA_ROOT': get_weave_data_root()
    }

    for key, value in replaces.items():
        text = text.replace(key, value)
    
    return text


def get_provider_provider():
    '''
    Returns the provider provider.

    For example,

    .. code-block:: python

        import axiomic.models as models
        models.get_provider_provider()

    '''
    return context.get_provider_provider()


import pytest
from unittest.mock import patch, mock_open
import json
import yaml

# Assuming your module is named 'my_module'
import axiomic.data.loaders as loaders

def test_read_unstructured_success():
    # Mock opening a file that is supposed to read text content
    mock_data = "this is my text"
    with patch("builtins.open", mock_open(read_data=mock_data)):
        # No exception means it passed
        assert loaders.read_text("file.txt") == "this is my text"


def test_read_structured_json_success():
    # Mock opening a file that is supposed to read text content
    mock_data = '["one", "two", "three"]'
    with patch("builtins.open", mock_open(read_data=mock_data)):
        # No exception means it passed
        assert loaders.read_structured("file.json") == ["one", "two", "three"]


def test_read_structured_yaml_success():
    # Mock opening a file that is supposed to read text content
    mock_data = yaml.dump(["one", "two", "three"])
    with patch("builtins.open", mock_open(read_data=mock_data)):
        # No exception means it passed
        assert loaders.read_structured("file.yaml") == ["one", "two", "three"]


def test_load_json_serialized():
    # Mock reading a valid list of strings
    mock_data = '[[1, 2, 3], [5]]'
    with patch("builtins.open", mock_open(read_data=mock_data)):
        # No exception means it passed
        assert list(loaders.load_serialized_jsons("file.json")) == ['[1, 2, 3]', '[5]']


def test_load_yaml_serialized_json():
    # Mock opening a file that is supposed to read text content
    mock_data = '''
-   - 1
    - 2
- two
- three
    '''
    with patch("builtins.open", mock_open(read_data=mock_data)):
        # No exception means it passed
        assert list(loaders.load_jsons_serialized("file.yaml")) == ["[1, 2]", '"two"', '"three"']


def test_load_chat_text_json():
    # Mock opening a file that is supposed to read text content
    mock_data = '''
[
    {"user": "Hello Agent!", "agent": "Hi Person!"},
    {"user": "What's 5 + 5?", "agent": "It's 10"}
]
    '''

    with patch("builtins.open", mock_open(read_data=mock_data)):
        # No exception means it passed
        assert loaders.load_chat_text("file.json").as_list() == [("Hello Agent!", "Hi Person!"), ("What's 5 + 5?", "It's 10")]


def test_load_chat_text_yaml():
    # Mock opening a file that is supposed to read text content
    mock_data = '''
        - user: Hello Agent!
          agent: Hi Person!
        - user: What's 5 + 5?
          agent: It's 10
    '''

    with patch("builtins.open", mock_open(read_data=mock_data)):
        # No exception means it passed
        assert loaders.load_chat_text("file.yaml").as_list() == [("Hello Agent!", "Hi Person!"), ("What's 5 + 5?", "It's 10")]


def test_load_chat_serialized_json_json():
    # Mock opening a file that is supposed to read text content
    mock_data = '''
[
    {"user": {"a": 1}, "agent": {"b": 2}},
    {"user": {"a": 2}, "agent": {"b": 3}}
]
    '''

    schema = '{"type": "object"}'

    with patch("builtins.open", mock_open(read_data=mock_data)):
        # No exception means it passed
        assert loaders.load_chat_json_serialized("file.json", user_schema=schema, agent_schema=schema).as_list() == [('{"a": 1}', '{"b": 2}'), ('{"a": 2}', '{"b": 3}')]
    





import axiomic.models.context as context
import axiomic.models.generic as generic

Generic = generic.Generic

class OpenAI:

    class Text:
        GPT3_5 = context.LLMConfig(
            llm_provider_name='openai_text',
            llm_model_name='gpt-3.5-turbo',
            llm_temperature=0.5,
            llm_max_tokens=1024,
            _context_name='OpenAI.text.GPT3_5'
        )

        GPT4 = context.LLMConfig(
            llm_provider_name='openai_text',
            llm_model_name='gpt-4',
            llm_temperature=0.5,
            llm_max_tokens=1024,
            _context_name='OpenAI.text.GPT4'
        )

    class MultiModal:
        GPT4o = context.LLMConfig(
            llm_provider_name='openai_text',
            llm_model_name='gpt-4o',
            llm_temperature=0.5,
            llm_max_tokens=4096,
            _context_name='OpenAI.MultiModal.Gpt4o'
        )

    class Image:
        Dalle3_1024x1024 = context.Config(
            _context_name='OpenAiDalle3_1024x1024',
            image_provider_name='openai_img',
            image_model_name='dall-e-3',
            image_width=1024,
            image_height=1024
        )

        Dalle3_1024x1792 = context.Config(
            _context_name='OpenAiDalle3_1024x1792',
            image_model_name='dall-e-3',
            image_width=1024,
            image_height=1792
        )

        Dalle3_1792x1024 = context.Config(
            _context_name='OpenAiDalle3_1792x1024',
            image_model_name='dall-e-3',
            image_width=1792,
            image_height=1024
        )

    def bind():
        generic.bind(Generic.Text.Small, OpenAI.Text.GPT3_5)
        generic.bind(Generic.Text.Medium, OpenAI.Text.GPT3_5)
        generic.bind(Generic.Text.Large, OpenAI.MultiModal.GPT4o)
        generic.bind(Generic.Image.Medium, OpenAI.Image.Dalle3_1024x1024)





import json
import typing
import jsonschema


def _to_schema_dict(schema: typing.Union[str, dict]):
    if isinstance(schema, dict):
        return schema
    return json.loads(schema)


class JsonSchema:
    
    def __init__(self, schema: typing.Union[str, dict]):
        self.schema = _to_schema_dict(schema)

    def validate(self, thing: any) -> bool:
        try:
            jsonschema.validate(thing, self.schema)
            return True
        except jsonschema.ValidationError:
            return False
        except jsonschema.SchemaError:
            return False

    def validate_serialized(self, thing: str) -> bool:
        try:
            jsonschema.validate(json.loads(thing), self.schema)
            return True
        except jsonschema.ValidationError:
            return False
        except jsonschema.SchemaError:
            return False





import axiomic.configure.runtime_config as runtime_config
import axiomic.models.context as context
import os

import axiomic.logalytics.sink as sink

Config = context.Config

class Generic:

    class Text:
            
        Small = Config(
            llm_provider_name='RUNTIME:small_text_provider',
            llm_model_name='RUNTIME:small_text_model',
            llm_temperature=0.5,
            llm_max_tokens=1024,
            _context_name='Generic.Text.Small'
        )
        ''' The default small LLM, e.g. Claude Haiku, which will be discovered based on env varibles and available API keys.  '''

        Medium = Config(
            llm_provider_name='RUNTIME:medium_text_provider',
            llm_model_name='RUNTIME:medium_text_model',
            llm_temperature=0.5,
            llm_max_tokens=1024,
            _context_name='Generic.Text.Medium'
        )
        ''' The default Medium LLM, e.g. GPT-3, Claude Sonnet, which will be discovered based on env varibles and available API keys.  '''

        Large = Config(
            llm_provider_name='RUNTIME:large_text_provider',
            llm_model_name='RUNTIME:large_text_model',
            llm_temperature=0.5,
            llm_max_tokens=1024,
            _context_name='Generic.Text.Large'
        )
        ''' The default Medium LLM, e.g. GPT-4, Claude Opus, which will be discovered based on env varibles and available API keys.  '''

    class Image:

        Medium = Config(
            image_provider_name='RUNTIME:medium_image_provider',
            image_model_name='RUNTIME:medium_image_model',
            image_width='RUNTIME:medium_image_width',
            image_height='RUNTIME:medium_image_height',
            _context_name='Generic.Image.Medium',
        )


def bind(generic_cfg, non_generic_cfg):
    '''
    Binds the generic configuration to a specific configuration.

    For exmaple,

    .. code-block:: python

        bind(Generic.Text.Small, quick_together.Together.Text.Llama.llama_3_70b_chat_hf)
    
    '''
    generic_cfg = generic_cfg.context
    non_generic_cfg = non_generic_cfg.context

    def maybe_set_attr(generic_cfg, non_generic_cfg, attr):
        if type(getattr(generic_cfg, attr)) != str:
            return

        if 'RUNTIME:' not in getattr(generic_cfg, attr):
            return

        runtime_attr_name = getattr(generic_cfg, attr).split(':')[-1]
        runtime_config.GLBOAL_RUNTIME_CONFIG.__setattr__(f'{runtime_attr_name}', getattr(non_generic_cfg, attr))

    for attr in dir(generic_cfg):
        if not attr.startswith('_'):
            maybe_set_attr(generic_cfg, non_generic_cfg, attr)

    # sink.discover_config(f'bind {generic_cfg._context_name} -> {non_generic_cfg._context_name}')


def bind_if_env_key(env_key, generic_cfg, non_generic_cfg):
    '''
    Binds the generic configuration to a specific configuration if the environment key is present.

    For exmaple,

    .. code-block:: python

        bind_if_env_key('TOGETHER_API_KEY', Generic.Text.Small, quick_together.Together.Text.Llama.llama_3_70b_chat_hf)
    
    '''
    if env_key in os.environ:
        bind(generic_cfg, non_generic_cfg)



class Binder:
    def __init__(self, env_key, bind_pairs):
        self.env_key = env_key
        self.bind_pairs = bind_pairs

    def has_env_key(self):
        return self.env_key in os.environ

    def bind(self):
        for generic_cfg, non_generic_cfg in self.bind_pairs:
            bind(generic_cfg, non_generic_cfg)



from setuptools import setup, find_packages


# Read the requirements from the requirements.txt file
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='axiomic',
    version='0.0.1',
    packages=find_packages(),
    author='Victor Bittorf',
    author_email='bitfort@gmail.com',
    description='Primitives for Genreative AI.',
    url='https://github.com/axiomic-ai/axiomic',
    install_requires=requirements,
    package_data={
        'axiomic': ['axiomic/protos/axiomic.proto'],
    },
    entry_points={
        'console_scripts': [
            'axiomic_params = aximoic.tools.params:main',
        ],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
)





import axiomic.protos as protos

import axiomic.providers.llm_provider.llm_provider as llm_provider
import axiomic.models as models

import re
import json

from icecream import ic 

import axiomic.configure.runtime_config as runtime_config

ic.configureOutput(includeContext=True)


def history_complete(node: protos.axiomic.HistoryCompleteNode, c, weave_node: protos.axiomic.AxiomicNode) -> str:
    cw = c.resolve_node(node.user_message)
    history_pairs = []

    for pair in node.chat_history:
        user_message = c.resolve_node(pair.user_message)
        agent_message = c.resolve_node(pair.assistant_message)
        history_pairs.append((user_message, agent_message))
    sp = c.resolve_node(node.system_prompt)

    params = node.complete_params

    llm_provider_name = params.llm_provider_name
    llm_model_name = params.llm_model_name
    if 'RUNTIME:' in llm_provider_name:
        llm_provider_name, llm_model_name = runtime_config.resolve_runtime_provider_and_model(llm_provider_name, llm_model_name)

    req = llm_provider.LlmHistoryInferenceRequest(
        llm_provider_name,
        llm_model_name,
        params.llm_temperature,
        params.llm_max_tokens,
        cw,
        sp,
        history_pairs
    )

    resp = models.get_provider_provider().get_provider(llm_provider_name).infer_history(req)


    # print('history_copmlete.py: ask_history: Inferring for ', weave_node.name, '  ::: ', node.user_message.name, 'with complexity', complexity, 'system prompt', node.system_prompt.name)
    # resp = ask_history(cw, history_pairs, sp, complexity_level=complexity)
    return resp.response



import threading

import axiomic.logalytics.events as events

import axiomic.logalytics.sink as sink


class EventContextManager:
    _thread_local = threading.local()  # Thread-local storage

    def __init__(self, name, event_type: events.EventType, fields: dict):
        self.name = name
        self.event_type = event_type
        self.fields = fields
        self.event = events.Event(name, event_type, fields)
        self.closed = False

    def __enter__(self):
        if not hasattr(EventContextManager._thread_local, 'stack'):
            EventContextManager._thread_local.stack = []  # Initialize stack if not present
        EventContextManager._thread_local.stack.append(
            self)  # Push the current context onto the stack
        sink.event_sink(self.event)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Pop the current context from the stack
        if exc_val is not None:
            EventContextManager._thread_local.stack.pop()
            return 
        if self.event_type in events.MAP_START_TO_END.keys() and not self.closed:
            raise Exception('InternalError: Failed to close event')
        
        if self.event_type == events.EventType.TASK_START:
            event = events.Event(self.name, events.EventType.TASK_SUCCESS, self.fields)
            sink.event_sink(event)
        EventContextManager._thread_local.stack.pop()
    
    def end(self, **kwargs):
        self.closed = True
        end_type = events.MAP_START_TO_END[self.event_type]
        event = events.Event(self.name, end_type, kwargs)
        sink.event_sink(event)

    def task_fail(self, fields: dict):
        event = events.Event(self.name, events.EventType.TASK_FAILURE, fields)
        sink.event_sink(events.Event(event))



def Task(name, fields: dict):
    return EventContextManager(name, events.EventType.TASK_START, fields)


def LlmInference(name, fields: dict):
    return EventContextManager(name, events.EventType.LLM_INFERENCE_START, fields)

def ImageInference(name, fields: dict):
    return EventContextManager(name, events.EventType.IMAGE_INFERENCE_START, fields)

def RagQuery(name, fields: dict):
    return EventContextManager(name, events.EventType.RAG_QUERY_START, fields)


def _global_event_chain():
    if not hasattr(EventContextManager._thread_local, 'stack'):
        return []
    return EventContextManager._thread_local.stack







import dataclasses
import os



@dataclasses.dataclass
class Config:
    enable_caching: bool = True
    enable_file_cache: bool = False 
    enable_mem_cache: bool = True
    thought_crime_retry_count: int = 3
    data_root_dir: str = './axiomic_data'

    def get_data_filepath(self, dirname, filename):
        # return os.path.join({global_config.data_root_dir}/{dirname}/{filename})
        return os.path.abspath(os.path.join(self.data_root_dir, dirname, filename))


def set_global_config(cfg: Config):
    global global_config
    global_config = cfg



global_config = Config()




import json
import pydantic


def coerce_to_str(thing: any) -> str:
    '''
    Converts any object to a string.

    Args:
        thing (any): The object to convert.

    Returns:
        str: The string representation of the object.
    '''

    if thing is None:
        return None

    if isinstance(thing, pydantic.BaseModel):
        return str(thing.json())

    if type(thing) is str:
        return thing

    # assume it should be serialized to json
    return json.dumps(thing)
